#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Utility functions for the test suite."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from pathlib import Path


def resource_check(*args):
    """Checks to see if the files exist. And returns a tuple.

       Using the first element as the truth value and the second
       as the test value in a ``self.assertEqual(truth, test)``
       provides a much more useful failure message than a bunch
       of ``self.assertTrue(os.path.isfile(filename))``
    """
    CheckReturn = collections.namedtuple("CheckReturn", ["truth", "test"])

    truth = "Missing Resources:"
    test = truth

    the_missing = []
    for path in map(Path, args):
        if not path.is_file():
            the_missing.append(path)
    if len(the_missing) > 0:
        test = "\n".join(map(lambda x: "    " + str(x), the_missing))
        test += "\n probably just need to 'make test-resources'"

    return CheckReturn(truth, test)



#!/usr/bin/env python
"""Generate radiometrically corrected HiRISE Channel products.

These are the functionalities that the EDR_Stats pipeline performs that are
reproduced here:

- Perform a radiometric calibration correction and conversion to "I/F" units
  using the ISIS ``hical`` program.
- Perform furrow correction. The image columns at the channel join
  are checked for furrows. Pixels in the furrow region (at the channel
  join) whose DN values have gone above a threshold are set to the
  null pixel value as these pixels can not be calibrated. If a
  RED-filter image experiences furrowing, a rare event usually caused
  by an improperly commanded observation, then the furrowed pixels
  will be permanently set to null pixels for all of the HiRISE standard
  and extras products. For BG and IR-filter images that make up the
  color products, the furrowed pixels will be interpolated in the
  HiColorNorm pipeline step.
- If an observation is determined to have furrows then an entry
  with the key "zapped" is set to "true" in the output .json file.
- Due to HiRISE instrument instability problems, the HiCal pipeline
  performs a noise reduction procedure to reduce the number of bad
  pixels in an image observation. The noise correction is applied
  when the standard deviation of the dark pixel or mask regions
  exceed a threshold, or if the number of LIS pixels exceeds a
  threshold.  A newer version of that processing can be engaged
  if --bitflipwidth is set to a non-zero value.
- Sometimes operating at lower temperatures can result in LIS pixels
  in the calibration areas, but not in the image area that can result
  in an erroneous calibration.  Some new processing to deal with that
  can be engaged if --lisfix is set to a value less than 1.
- A high-pass filter "cubenorm" step is applied to the calibrated
  image. Due to camera instabilities, residual vertical striping
  often exists in the imaging that is corrected by this empirical
  method. The average standard deviation of this change is calculated
  and stored in the .json file.
- The ISIS program "hidestripe" is applied to suppress horizontal
  stripping seen whenever an observation is acquired using mixed-binning
  commanding. The standard deviation of this change is calculated
  and stored in the .json file.


HiCal is the second step in the HiRISE processing chain, after EDR_Stats and
results in data that is read for the next step: HiStitch.  If you are
running HiCal individually without all EDRs from the Observation in the
same directory, please be aware of the --bin2 and --bin4 options.


Data Flow
---------
Input Products:

- A ``.cub`` file that has been converted from an EDR.
- A ``.json`` file that contains summary information about the .cub file.

Output Products:

- A calibrated ``.cub`` file for each input ``.cub`` file processed.
- Modifies each provided ``.json`` file.

"""

# Copyright 2004-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiCal version 4.1.1 (2021/05/03),
# and on the Perl HiCal program ($Revision: 1.58 $
#                                $Date: 2021/05/03 22:33:55 $)
# and on the Perl NoiseFilter program ($Revision: 1.19
#                                      $Date: 2021/04/22 22:38:39 $)
# by Eric Eliason and Richard Leis as employees of the University of Arizona.

import argparse
import collections
import concurrent.futures
import copy
import csv
import json
import logging
import math
import os
import pkg_resources
import re
import statistics
import sys
import textwrap
import warnings
from datetime import datetime
from pathlib import Path

import numpy as np

import kalasiris as isis
import kalasiris.version as isisversion
import pvl.new as pvl

import hiproc.bitflips as bf
import hiproc.lisfix as lisfix
import hiproc.hirise as hirise
import hiproc.util as util

logger = logging.getLogger(__name__)

tempfile_re = re.compile(r"HiCal-\d{12}\.")


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("-o", "--output", required=False, default=".HiCal.cub")
    parser.add_argument(
        "-c",
        "--conf",
        required=False,
        type=Path,
        default=Path(
            pkg_resources.resource_filename(
                __name__,
                'data/HiCal.conf'
            )
        ),
        help="Configuration file to use, Defaults to "
             "HiCal.conf distributed with the library."
    )
    parser.add_argument(
        "--db",
        required=False,
        default=".HiCat.json",
        help="The .json file to use.  Optionally, if it "
        "starts with a '.' it is considered an extension "
        "and will be swapped with the input file's "
        "extension to find the .json file to use. "
        "Default: %(default)s",
    )
    parser.add_argument(
        "-b",
        "--bitflipwidth",
        required=False,
        type=int,
        default=None,
        help="The number of medstd widths for bit-flip cleaning. Leaving it "
             "unset will result in the setting being determined by the value "
             "of LIS_Mask in the configuration file.  Explicitly setting it to "
             "zero implies no bit-flip cleaning and will use the old "
             "ANALYZECUBENORMSTATS behavior. Explicitly setting it to 5 is "
             "identical to the BITFLIPS setting.",
    )
    parser.add_argument(
        "-t",
        "--lis_tolerance",
        type=float,
        default=None,
        help="If the fraction of LIS pixels in each of the Reverse-Clock and "
             "Buffer areas are higher than this value, lis-fix processing "
             "will occur for that area. Leaving it unset will result in "
             "the value being determined by the value of LIS_Fix in the "
             "configuration fie.  Explicitly setting it to 0.4 is the same "
             "as LIS_Fix = TRUE, and explicitly setting it to 1.0 is the "
             "same as LIS_Fix = FALSE."
    )
    # parser.add_argument('--hgfconf', required=False, default='HiGainFx.conf')
    parser.add_argument(
        "--nfconf",
        required=False,
        type=Path,
        default=Path(
            pkg_resources.resource_filename(
                __name__,
                'data/NoiseFilter.conf'
            )
        ),
        help="Defaults to NoiseFilter.conf distributed with the library."
    )
    parser.add_argument(
        "--bin2",
        required=False,
        action="store_true",
        default=None,
        help="By default HiCal will look at all of the HiRISE .cub files in "
             "the directory with the input file(s) in it, and determine "
             "whether any of the EDRs from this observation was bin2 and will "
             "internally take care of setting this switch, but if you're just "
             "processing some RED channels and not all of them, then although "
             "IR10 may be binned, and you didn't convert it to a .cub file or "
             "even download it, then it isn't there for this program to find, "
             "and explicitly setting this switch is needed."
    )
    parser.add_argument(
        "--nobin2",
        required=False,
        action="store_false",
        default=None,
        dest="bin2",
        help="Forcibly sets the --bin2 switch to off, even if it would "
             "otherwise be automatically determined to be on."
    )
    parser.add_argument(
        "--bin4",
        required=False,
        action="store_true",
        default=None,
        help="Same as the --bin2 switch, but for setting whether any EDRs "
             "from this observation are bin4."
    )
    parser.add_argument(
        "--nobin4",
        required=False,
        action="store_false",
        default=None,
        dest="bin4",
        help="Forcibly sets the --bin4 switch to off, even if it would "
             "otherwise be automatically determined to be on."
    )
    parser.add_argument(
        "-n", "--info",
        action="store_true",
        help="If given, the program will not run, but will instead report "
             "information about the input files and various things that "
             "the program would have done, if it was run regularly."
    )
    parser.add_argument(
        "--max_workers",
        default=None,
        type=int,
        help="If more than one image is provided to work on, this program "
             "will engage multiprocessing to parallelize the work.  This "
             "multiprocessing will default to the number of processors on the "
             "machine.  If you want to throttle this to use less resources on "
             "your machin, indicate the number of processors you want to use."
    )
    parser.add_argument(
        "cube",
        metavar="cube_file(s)",
        nargs="+",
        help="More than one can be listed here.",
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    if len(args.cube) > 1 and (
        not args.output.startswith(".") or not args.db.startswith(".")
    ):
        logger.critical(
            "With more than one input cube file, the --output "
            "and --db must start with a period, and one of them "
            f"does not: {args.output} {args.db}"
        )
        sys.exit()

    try:
        conf = conf_setup(pvl.load(args.conf), pvl.load(args.nfconf))
    except (TypeError, NotADirectoryError, FileNotFoundError) as err:
        logger.critical(err)
        sys.exit()

    if args.bitflipwidth is None:
        d = {"ANALYZECUBENORMSTATS": 0, "BITFLIPS": 5}
        args.bitflipwidth = d[conf["HiCal"]["LIS_Mask"]]

    if args.lis_tolerance is None:
        d = {True: 0.4, False: 1.0}
        args.lis_tolerance = d[conf["HiCal"]["LIS_Fix"]]

    with util.main_exceptions(args.verbose, logger=logger):
        if args.info:
            for c in args.cube:
                db_path = util.pid_path_w_suffix(args.db, c)
                with open(db_path, "r") as f:
                    db = json.load(f)

                if args.info:
                    print(f"HiCal Information for {c}:")
                    print(textwrap.indent(
                        "\n".join(
                            norun_info(c, db, conf, args.bin2, args.bin4)
                        ),
                        "  "
                    ))
            return

        if len(args.cube) == 1:
            # No need to fire up the multiprocessing for just one image.
            db_path = util.pid_path_w_suffix(args.db, args.cube[0])
            with open(db_path, "r") as f:
                db = json.load(f)

            out_cube = util.path_w_suffix(args.output, args.cube[0])
            db = HiCal(
                args.cube[0],
                out_cube,
                db,
                conf,
                args.conf,
                args.bin2,
                args.bin4,
                args.bitflipwidth,
                args.lis_tolerance,
                keep=args.keep,
            )
            with open(db_path, "w") as f:
                json.dump(db, f, indent=0, sort_keys=True)

            logger.info(f"Wrote {db_path}")
        else:
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=args.max_workers
            ) as executor:
                future_dbs = dict()
                for c in args.cube:
                    db_path = util.pid_path_w_suffix(args.db, c)
                    with open(db_path, "r") as f:
                        db = json.load(f)

                    out_cube = util.path_w_suffix(args.output, c)

                    f = executor.submit(
                        HiCal,
                        c,
                        out_cube,
                        db,
                        conf,
                        args.conf,
                        args.bin2,
                        args.bin4,
                        args.bitflipwidth,
                        args.lis_tolerance,
                        keep=args.keep,
                    )
                    future_dbs[f] = db_path

                for future in concurrent.futures.as_completed(future_dbs):
                    db_path = future_dbs[future]
                    with open(db_path, "w") as f:
                        json.dump(future.result(), f, indent=0, sort_keys=True)

                    logger.info(f"Wrote {db_path}")
    return


def conf_setup(conf: dict, nfconf: dict) -> dict:
    # Get Configuration Parameters
    conf_check(conf)

    # # If the sub-conf arguments aren't 'findable', look for them in
    # # the main conf directory.
    # # hgf_path = util.get_path(Path(args.hgfconf), Path(args.conf).parent)
    # nf_path = util.get_path(
    #     Path(nfconf_path),
    #     (
    #         Path(conf_path).parent,
    #         Path(__file__).resolve().parent.parent / "data",
    #     ),
    # )

    # Merge the configuration files together into a single dict
    # conf['HiGainFx'] = pvl.load(str(hgf_path))['HiGainFx']
    # conf["NoiseFilter"] = pvl.load(str(nf_path))["NoiseFilter"]
    conf["NoiseFilter"] = nfconf["NoiseFilter"]

    return conf


def norun_info(cube: os.PathLike, db: dict, conf: dict, bin2: bool, bin4: bool):

    lines = list()
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("ignore")
        in_cube, cid, ccdchan, lis_per = setup(cube, db, conf)
        if len(w) > 0:
            for wa in w:
                lines.append(f"Would have warned: {wa.message}")

    lines.append(
        f'Image LIS: {db["LOW_SATURATED_PIXELS"]} pixels, '
        f'{lis_per:.2f}%'
    )

    # Non-imaging areas:
    label = pvl.load(cube)
    specialpix = getattr(
        isis.specialpixels, label["IsisCube"]["Core"]["Pixels"]["Type"]
    )
    t_name = "HiRISE Calibration Image"
    hci_dict = isis.cube.get_table(in_cube, t_name)
    cal_vals = np.array(hci_dict["Calibration"])
    cal_image = np.ma.masked_outside(
        cal_vals, specialpix.Min, specialpix.Max
    )
    revclk_liscount = np.ma.count_masked(cal_image[:20, :])
    revclk_lisfrac = revclk_liscount / cal_image[:20, :].size
    lines.append(
        f"Reverse-Clock LIS: {revclk_liscount} pixels, "
        f"{revclk_lisfrac:.2%}"
    )

    hca_dict = isis.cube.get_table(cube, "HiRISE Calibration Ancillary")
    calbuf_vals = np.array(hca_dict["BufferPixels"])
    calbuf = np.ma.masked_outside(
        calbuf_vals, specialpix.Min, specialpix.Max
    )
    calbuf_liscount = np.ma.count_masked(calbuf[:20, :])
    calbuf_lisfrac = calbuf_liscount / calbuf[:20, :].size
    lines.append(
        f"Reverse-Clock Buffer LIS: {calbuf_liscount} pixels, "
        f"{calbuf_lisfrac:.2%}"
    )

    ha_dict = isis.cube.get_table(cube, "HiRISE Ancillary")
    buffer_vals = np.array(ha_dict["BufferPixels"])
    buffer = np.ma.masked_outside(
        buffer_vals, specialpix.Min, specialpix.Max
    )
    buffer_liscount = np.ma.count_masked(buffer)
    buffer_lisfrac = buffer_liscount / buffer.size
    lines.append(
        f"Buffer LIS: {buffer_liscount} pixels, "
        f"{buffer_lisfrac:.2%}"
    )

    b = 1, 2, 4, 8, 16
    flags = set_flags(
        conf["HiCal"], db, ccdchan, b.index(int(db["BINNING"]))
    )
    lines.append(str(flags))

    destripe_filter = check_destripe(in_cube, int(db["BINNING"]), bin2, bin4)
    lines.append(f"Would the destripe filter be run: {destripe_filter}")

    return lines


def setup(cube: os.PathLike, db: dict, conf: dict):
    in_cube = Path(cube)
    cid = hirise.get_ChannelID_fromfile(in_cube)
    if str(cid) != db["PRODUCT_ID"]:
        raise ValueError(
            "The Product ID in the file ({}) does not match "
            "the one in the database ({}).".format(str(cid), db["PRODUCT_ID"])
        )
    ccdchan = (cid.get_ccd(), cid.channel)
    lis_per = (
        float(db["LOW_SATURATED_PIXELS"])
        / (int(db["IMAGE_LINES"]) * int(db["LINE_SAMPLES"]))
        * 100.0
    )
    # if (
    #     ccdchan == ("IR10", "1")
    #     and lis_per > conf["HiCal"]["HiCal_Bypass_IR10_1"]
    # ):
    #     raise UserWarning("Bypassing IR10_1.")
    if lis_per > conf["HiCal"]["HiCal_Bypass"]:
        raise UserWarning(
            f"The image area has a LIS percent ({lis_per}) greater than "
            f"{conf['HiCal']['HiCal_Bypass']}, so this image will not be "
            f"processed through HiCal."
        )

    return in_cube, cid, ccdchan, lis_per


def HiCal(
    in_cube: os.PathLike,
    out_cube: os.PathLike,
    db: dict,
    conf: dict,
    conf_path: os.PathLike,
    bin2: bool,
    bin4: bool,
    bitflipwidth=0,
    lis_tolerance=1.0,
    keep=False,
) -> dict:
    # The original Perl Setup00() builds data structures that we don't
    # need.
    # The original Perl Setup01() set up data routing and did filename
    # checking that we don't need here.

    # The original Perl Setup03 queried the HiCat.Planned_Observations
    # table to get all of the possible binning values for CCDs in this
    # CCD's observation.  We now accomplish the same thing with
    # check_destripe() farther down.

    # The original Perl ProcessingSwitches() set various flags.  That
    # functionality is now broken up and spread out to the check_destripe()
    # function, the set_flags() function, and the if statement below that
    # checks HiCal_Bypass_IR10_1
    in_cube, cid, ccdchan, lis_per = setup(in_cube, db, conf)
    logger.info(f"{cid}: HiCal start: {in_cube}")

    destripe = check_destripe(in_cube, int(db["BINNING"]), bin2, bin4)

    # The original Perl SetHiCalVersion() placed the HiCal version into the
    # db, but since we're not interested in persistence, we can ignore it.

    # Allows for indexing in lists ordered by bin value.
    b = 1, 2, 4, 8, 16

    in_c = Path(in_cube)

    # Keep from having to write out ['HiCal'] all the time.
    hconf = conf["HiCal"]

    # This string will get placed in the filename for all of our
    # temporary files. It will (hopefully) prevent collisions with
    # existing files and also allow for easy clean-up if keep=True
    temp_token = datetime.now().strftime("HiCal-%y%m%d%H%M%S")

    logger.info(f"{cid}: destripe: {destripe}")

    flags = set_flags(
        hconf, db, ccdchan, b.index(int(db["BINNING"])), bitflipwidth > 0
    )
    logger.info(f"{cid}: {flags}")

    # Start processing cube files
    to_delete = isis.PathSet()
    furrows_found = False
    if int(db["BINNING"]) > 1 and float(db["IMAGE_MEAN"]) > 7000.0:
        furrow_cube = to_delete.add(
            in_c.with_suffix(f".{temp_token}.ffix.cub")
        )
        furrows_found = furrow_nulling(
            in_c, furrow_cube, int(db["BINNING"]), ccdchan
        )
        next_cube = furrow_cube
    else:
        next_cube = to_delete.add(in_c.with_suffix(f".{temp_token}.cub"))
        logger.info(f"{cid}: Symlink {in_c} to {next_cube}")
        next_cube.symlink_to(in_c.resolve())
    logger.info(f"{cid}: Furrow Fix application: {furrows_found}")

    # Run hical
    lis_per = (
        int(db["LOW_SATURATED_PIXELS"])
        / (int(db["IMAGE_LINES"]) * int(db["LINE_SAMPLES"]))
        * 100.0
    )
    hical_file = to_delete.add(next_cube.with_suffix(".hical.cub"))
    hical_status = run_hical(
        next_cube,
        hical_file,
        conf,
        conf_path,
        lis_per,
        float(db["IMAGE_BUFFER_MEAN"]),
        int(db["BINNING"]),
        flags.noise_filter,
        bitflipwidth=bitflipwidth,
        lis_tolerance=lis_tolerance,
        keep=keep,
    )
    next_cube = hical_file

    if furrows_found:
        lpfz_file = to_delete.add(next_cube.with_suffix(".lpfz.cub"))
        isis.lowpass(
            next_cube,
            to=lpfz_file,
            lines=3,
            samples=3,
            minopt="COUNT",
            minimum=5,
            filter_="OUTSIDE",
        )
        next_cube = lpfz_file

    # # Perform gain-drift correction
    # if(db['BINNING'] != '8'):  # There is no gain fix for bin8 imaging
    #     higain_file = to_delete.add(next_cube.with_suffix('.fx.cub'))
    #     hgf_path = util.get_path(Path('HiGainFx.conf'),
    #                              Path(conf_path).parent)
    #     conf['HiGainFx'] = pvl.load(str(hgf_path))['HiGainFx']
    #     HiGainFx(next_cube, higain_file,
    #              conf['HiGainFx']['HiGainFx_Coefficient_Path'],
    #              conf['HiGainFx']['HiGainFx_Version'])
    #     next_cube = higain_file

    # Perform the high-pass filter cubenorm steps

    (sl, nl) = set_lines(
        int(hconf["HiCal_Bin1_Skip_Top_Lines"]),
        int(hconf["HiCal_Bin1_Skip_Bot_Lines"]),
        int(db["BINNING"]),
        int(db["IMAGE_LINES"]),
    )
    crop_file = to_delete.add(next_cube.with_suffix(".crop.cub"))
    isis.crop(next_cube, to=crop_file, line=sl, nlines=nl)

    # In the original Perl, this file was kept for the next step, HiStitch.
    # However, since FurrowCheck() is now performed here, it does not need
    # to be kept.
    stats_file = to_delete.add(out_cube.with_suffix(".cubenorm.tab"))
    stats_fix_file = to_delete.add(next_cube.with_suffix(".cubenorm_fix.tab"))
    isis.cubenorm(crop_file, stats=stats_file, format_="TABLE")

    (std_final, zapped) = Cubenorm_Filter(
        stats_file,
        stats_fix_file,
        boxfilter=5,
        pause=True,
        divide=flags.divide,
        chan=int(ccdchan[1]),
    )

    # Now perform the cubnorm_plus correction
    div_or_sub = {True: "DIVIDE", False: "SUBTRACT"}
    cubenorm_args = {
        "direction": "COLUMN",
        "statsource": "TABLE",
        "normalize": "AVERAGE",
        "preserve": "FALSE",
        "mode": div_or_sub[flags.divide],
    }
    cubenorm_file = to_delete.add(next_cube.with_suffix(".cn.cub"))
    to_s = "{}+SignedWord+{}:{}".format(
        cubenorm_file,
        hconf["HiCal_Normalization_Minimum"],
        hconf["HiCal_Normalization_Maximum"],
    )
    isis.cubenorm(
        next_cube, fromstats=stats_fix_file, to=to_s, **cubenorm_args
    )
    next_cube = cubenorm_file

    # NoiseFilter() - external Noise_Filter
    if flags.noise_filter:
        noisefilter_file = to_delete.add(next_cube.with_suffix(".nf.cub"))
        NoiseFilter(
            next_cube,
            output=noisefilter_file,
            conf=conf["NoiseFilter"],
            minimum=hconf["HiCal_Normalization_Minimum"],
            maximum=hconf["HiCal_Normalization_Maximum"],
            zapc=flags.zapcols,
            keep=keep,
        )
        next_cube = noisefilter_file

    # Hidestripe() - isis.[hidestripe,hipass,lowpass,algebra]
    diff_std_dev = None
    if destripe:
        hidestripe_file = to_delete.add(next_cube.with_suffix(".hd.cub"))
        diff_std_dev = Hidestripe(
            next_cube,
            hidestripe_file,
            int(db["BINNING"]),
            hconf["HiCal_Normalization_Minimum"],
            hconf["HiCal_Normalization_Maximum"],
            hconf["HiCal_Hidestripe_Correction"],
            int(db["LINE_SAMPLES"]),
            keep=keep,
        )
        next_cube = hidestripe_file

    # Create final output file
    to_delete.remove(next_cube)
    logger.info(f"{cid}: Rename {next_cube} to {out_cube}.")
    next_cube.rename(out_cube)

    if not keep:
        to_delete.unlink()

    db["HIGH_PASS_FILTER_CORRECTION_STANDARD_DEVIATION"] = std_final
    db["DESTRIPED_DIFFERENCE_STANDARD_DEVIATION"] = diff_std_dev
    db["zapped"] = zapped
    db["hical_status"] = hical_status
    # The zapped flag is not in the original code, I'm putting it in
    # to the DB to use in a later step.

    logger.info(f"{cid}: HiCal done: {out_cube}")
    return db


def FurrowCheck(vpnts: list, channel: int) -> bool:
    # This function was brought forward from HiStitch, because
    # it is more appropriate to perform the check here, and then
    # store the result for later.
    zap = False
    i = -1 * channel

    if vpnts[i] != max(vpnts):
        zap = True

    return zap


def conf_check(conf: dict) -> None:
    """Various checks on parameters in the configuration."""

    util.conf_check_strings(
        "HiCal_Clean_EDR_Stats",
        ("DELETE", "KEEP"),
        conf["HiCal"]["HiCal_Clean_EDR_Stats"],
    )

    util.conf_check_strings(
        "HiCal_HPF_Cubenorm",
        ("DIVIDE", "SUBTRACT"),
        conf["HiCal"]["HiCal_HPF_Cubenorm"],
    )

    util.conf_check_count(
        "HiCal_Noise_Processing",
        28,
        "CCD/channel",
        conf["HiCal"]["HiCal_Noise_Processing"],
    )

    util.conf_check_bounds(
        "HiCal_Bypass",
        (0.0, 100.0),
        conf["HiCal"]["HiCal_Bypass"],
    )

    # util.conf_check_bounds(
    #     "HiCal_Bypass_IR10_1",
    #     (0.0, 100.0),
    #     conf["HiCal"]["HiCal_Bypass_IR10_1"],
    # )

    util.conf_check_count(
        "HiCal_Noise_Bin_DarkPixel_STD",
        5,
        "possible bin mode",
        conf["HiCal"]["HiCal_Noise_Bin_DarkPixel_STD"],
    )

    util.conf_check_count(
        "HiCal_Noise_Bin_Mask_STD",
        5,
        "possible bin mode",
        conf["HiCal"]["HiCal_Noise_Bin_Mask_STD"],
    )

    util.conf_check_bounds(
        "HiCal_Normalization_Minimum",
        (-16384.0, 16364.0),
        conf["HiCal"]["HiCal_Normalization_Minimum"],
    )

    util.conf_check_bounds(
        "HiCal_Normalization_Maximum",
        (-16384.0, 16384.0),
        conf["HiCal"]["HiCal_Normalization_Maximum"],
    )

    util.conf_check_strings(
        "HiCal_Hidestripe_Correction",
        ("ADD", "MULTIPLE"),
        conf["HiCal"]["HiCal_Hidestripe_Correction"],
    )

    util.conf_check_bounds(
        "HiCal_Minimum_Percent",
        (0.0001, 2.0),
        conf["HiCal"]["HiCal_Minimum_Percent"],
    )

    util.conf_check_bounds(
        "HiCal_Maximum_Percent",
        (98.0, 99.9999),
        conf["HiCal"]["HiCal_Maximum_Percent"],
    )

    # There are checks for HiCal_Jpeg_Quality, HiCal_Thumb_Samples, and
    # HiCal_Browse_Samples in the original Perl that aren't relevant here,
    # ignoring.  There are also empty string checks, but if there's nothing
    # in a parameter, it will error when it gets used in the code, so no need
    # to check here.
    return


def check_destripe(
    cube: os.PathLike, mybinning: int, bin2=None, bin4=None
) -> bool:
    """Determines whether to run the destripe filtering based on whether *any*
    of the CCDs in an observation were bin 2 or bin 4.

    Without a database, we must either rely on the incoming bin2 and bin4
    arguments, or scour the directory that the input cube file is in, and
    read the cube labels present to try and find the information.
    """
    destripe_filter = False

    if bin2 is None and mybinning == 2:
        bin2 = True

    if bin4 is None and mybinning == 4:
        bin4 = True

    warn_message = None
    if (mybinning == 1 or mybinning == 2) and (bin2 is None or bin4 is None):
        binnings = get_bins_fromfiles(cube)
        powered_count = len(
            list(
                filter(
                    lambda x: x == "On",
                    isis.getkey_k(cube, "Instrument", "PoweredCpmmFlag").split(
                        ","
                    ),
                )
            )
        )
        if bin2 is None and "2" in binnings.values():
            bin2 = True

        if bin4 is None and "4" in binnings.values():
            bin4 = True

        if len(binnings) < powered_count:
            warn_message = (
                f"Expecting to find {powered_count} CCD files, but found"
                f"{len(binnings)}, as a consequence, cannot determine if the "
                "missing CCDs might have been bin 2 or bin 4.  "
                f"hidestripe will not be applied to this image ({cube})."
            )
            warn_message.format(powered_count, len(binnings))

    bin2 = bool(bin2)
    bin4 = bool(bin4)

    if (1 == mybinning and (bin2 or bin4)) or (2 == mybinning and bin4):
        destripe_filter = True
    elif warn_message is not None:
        logger.warning(warn_message)

    return destripe_filter


def get_bins_fromfiles(cube: os.PathLike) -> dict:
    """Extract summing values from all cubes in the input cube's directory
    that belong to the same Observation.
    """
    bins = dict()
    oid = hirise.get_ObsID_fromfile(cube)
    for path in Path(cube).parent.glob("*.cub"):
        if tempfile_re.search(str(path)) is not None:
            # Want to exclude files like "HiCal-%y%m%d%H%M%S"
            # which during multiprocessing could come and go between
            # the call to glob() above and the call to getkey_k()
            # below.
            continue
        try:
            p = hirise.get_ChannelID_fromfile(path)
            if (oid.phase, oid.orbit_number, oid.target) == (
                p.phase,
                p.orbit_number,
                p.target,
            ):
                ccd = p.get_ccd()
                if ccd in bins:
                    continue
                else:
                    bins[ccd] = isis.getkey_k(path, "Instrument", "Summing")
        except ValueError:
            # Couldn't extract a Product ID from a cube, ignore.
            continue
        except AttributeError:
            # Product ID doesn't have a CCD, ignore.
            continue
    return bins


def set_flags(
    conf: dict, db: dict, ccdchan: tuple, bindex: int, newlogic=True
) -> collections.namedtuple:
    """Set various processing flags based on various configuration
    parameters."""
    HiCalFlags = collections.namedtuple(
        "HiCalFlags", ["noise_filter", "zapcols", "divide"]
    )

    noise_filter = False
    if (process_this(ccdchan, conf["HiCal_Noise_Processing"])) and (
        (
            float(db["IMAGE_DARK_STANDARD_DEVIATION"])
            >= float(conf["HiCal_Noise_Bin_DarkPixel_STD"][bindex])
        )
        or (
            float(db["CAL_MASK_STANDARD_DEVIATION"])
            >= float(conf["HiCal_Noise_Bin_Mask_STD"][bindex])
        )
        or (
            int(db["LOW_SATURATED_PIXELS"])
            >= int(conf["HiCal_Noise_LIS_Count"])
        )
    ):
        noise_filter = True

    zapcols = False
    if int(db["LOW_SATURATED_PIXELS"]) >= int(conf["HiCal_Noise_LIS_Count"]):
        zapcols = True
        if newlogic:
            noise_filter = True

    if "DIVIDE" == conf["HiCal_HPF_Cubenorm"]:
        divide = True
    elif "SUBTRACT" == conf["HiCal_HPF_Cubenorm"]:
        divide = False
    else:
        raise KeyError(
            "The HiCal_HPF_Cubenorm keyword can be DIVIDE or "
            "SUBTRACT, but was "
            "{}".format(conf["HiCal_HPF_Cubenorm"])
        )

    return HiCalFlags(noise_filter, zapcols, divide)


def set_lines(
    skip_top: int, skip_bottom: int, binning: int, image_lines: int
) -> collections.namedtuple:
    """Determine the right values for the start lines and number of lines."""
    Lines = collections.namedtuple("Lines", ["start", "number"])

    sl = skip_top / binning + 1
    nl = image_lines - (skip_top + skip_bottom) / binning
    if nl < 2000 / binning:
        sl = 1
        nl = image_lines

    # The original Perl code doesn't force sl and nl back into ints, but we do.
    return Lines(int(sl), int(nl))


def run_hical(
    in_cube: os.PathLike,
    hical_cub: os.PathLike,
    conf: dict,
    conf_path: os.PathLike,
    lis_per: float,
    image_buffer_mean: float,
    binning: int,
    noiseclean: bool,
    bitflipwidth=0,
    lis_tolerance=1.0,
    keep=False,
) -> str:

    to_d = isis.PathSet()
    next_cube = Path(in_cube)
    status = "Standard"

    to_s = "{}+SignedWord+{}:{}".format(
        hical_cub,
        conf["HiCal"]["HiCal_Normalization_Minimum"],
        conf["HiCal"]["HiCal_Normalization_Maximum"],
    )
    hical_args = {"to": to_s, "units": "IOF"}
    if conf["HiCal"]["HiCal_ISIS_Conf"] != "DEFAULT":
        dirs = (
            Path(conf_path).parent,
            Path(
                pkg_resources.resource_filename(
                    __name__,
                    'data/'
                )
            )
        )
        if lis_per < 5 and image_buffer_mean > 0:
            hical_args["conf"] = util.get_path(
                conf["HiCal"]["HiCal_ISIS_Conf"], dirs
            )
        else:
            hical_args["conf"] = util.get_path(
                conf["HiCal"]["HiCal_ISIS_Conf_Noise"], dirs
            )
            status = "BadCal"

    if noiseclean:
        mask_cube = to_d.add(next_cube.with_suffix(".mask.cub"))
        mask(
            next_cube,
            mask_cube,
            conf["NoiseFilter"]["NoiseFilter_Raw_Min"],
            conf["NoiseFilter"]["NoiseFilter_Raw_Max"],
            binning,
            width=bitflipwidth,
            keep=keep,
        )
        next_cube = mask_cube

    lisfix_cube = next_cube.with_suffix(".lisfix.cub")
    if (
        lisfix.fix(next_cube, lisfix_cube, lis_tolerance)
    ):
        to_d.add(lisfix_cube)
        next_cube = lisfix_cube

    isis.hical(next_cube, **hical_args)

    if not keep:
        to_d.unlink()
        cid = hirise.get_ChannelID_fromfile(hical_cub)
        Path(hical_cub).with_name(str(cid)).with_suffix(".hical.log").unlink()

    return status


def process_this(ccdchan: tuple, flag_list: list) -> int:
    if len(flag_list) != 28:
        raise IndexError("The list must have 28 elements")
    ccd_number = int(hirise.get_ccdnumber(ccdchan[0]))
    i = (2 * ccd_number) + int(ccdchan[1])
    return int(flag_list[i])


def furrow_nulling(
    cube: os.PathLike,
    out_cube: os.PathLike,
    binning: int,
    ccdchan: tuple,
    keep=False,
) -> bool:
    furrows_found = False

    out_path = Path(out_cube)

    to_del = isis.PathSet()
    fcrop_file = to_del.add(out_path.with_suffix(".fcrop.cub"))

    isis.mask(cube, mask=cube, to=out_path, minimum=1000000, maximum=1000000)

    furrow_values = furrow_setup(ccdchan[0], binning)
    chan_samp = chan_samp_setup(int(ccdchan[1]), binning)

    # Crop out the portion of the image that will not be furrow checked.
    # ^- that's what the original said, but this is really cropping out
    #    the part that will be furrow corrected.
    isis.crop(cube, to=fcrop_file, samp=chan_samp.ssamp, nsamp=chan_samp.nsamp)

    isis.handmos(
        fcrop_file,
        mosaic=out_path,
        insamp=1,
        outsamp=chan_samp.ssamp,
        create="no",
    )

    # for each column subject to furrowing, crop out each column,
    # run mask to null pixels above the furrow threshold, and
    # mosaic into the output image.
    for (s, furrow_v) in zip(chan_samp.samp, furrow_values):
        fscrop_file = to_del.add(out_path.with_suffix(f".crop{s}.cub"))
        isis.crop(cube, to=fscrop_file, samp=s, nsamp=1)

        fsmask_file = to_del.add(out_path.with_suffix(f".mask{s}.cub"))
        isis.mask(
            fscrop_file,
            mask=fscrop_file,
            to=fsmask_file,
            minimum=0,
            maximum=furrow_v,
        )

        isis.handmos(fsmask_file, mosaic=out_path, insamp=1, outsamp=s)

    else:
        # finally, check to see if any furrow pixels were zapped.
        # The original Perl code checked on the fist iteration, this
        # mechanism checks on the last, not sure if there was anything
        # magical about the first, or maybe we should do it each time?
        fscrop_stats = pvl.loads(isis.stats(fscrop_file).stdout)
        fsmask_stats = pvl.loads(isis.stats(fsmask_file).stdout)

        if int(fsmask_stats["Results"]["NullPixels"]) > int(
            fscrop_stats["Results"]["NullPixels"]
        ):
            furrows_found = True
            # Perform simple 3x3 LPFZ filter to clean up the edges
            # along the furrow
            trim_file = out_path.with_suffix(".trim.cub")
            isis.trimfilter(
                out_path, to=trim_file, lines=3, samples=3, minopt="COUNT"
            )
            trim_file.rename(out_path)

    if not keep:
        to_del.unlink()

    return furrows_found


def mask(
    in_cube: os.PathLike,
    out_cube: os.PathLike,
    noisefilter_min: float,
    noisefilter_max: float,
    binning: int,
    width=5,
    cid=None,
    keep=False,
) -> None:
    """mask out unwanted pixels"""
    logger.debug(mask.__doc__)
    to_del = isis.PathSet()
    out_path = Path(out_cube)
    temp_cube = to_del.add(out_path.with_suffix(".mask_temp.cub"))

    if cid is None:
        cid = hirise.get_ChannelID_fromfile(in_cube)

    isis.mask(
        in_cube,
        mask=in_cube,
        to=temp_cube,
        minimum=noisefilter_min,
        maximum=noisefilter_max,
        preserve="INSIDE",
        spixels="NONE",
    )
    if width == 0:
        cubenorm_stats_file = to_del.add(temp_cube.with_suffix(".cn.stats"))
        isis.cubenorm(temp_cube, stats=cubenorm_stats_file)
        (mindn, maxdn) = analyze_cubenorm_stats(
            cubenorm_stats_file, binning, cid=cid
        )
        isis.mask(
            temp_cube,
            mask=temp_cube,
            to=out_path,
            minimum=mindn,
            maximum=maxdn,
            preserve="INSIDE",
            spixels="NONE",
        )
    else:
        # The analyze_cubenorm_stats() function is meant to make sure we
        # don't blow away valid data, with the philosophy that it is better to
        # let in a little bad data in order to keep the good.  However, in
        # images with a high percentage of LIS%, there is so much bad that it
        # swamps the good, so *this* approach attempts to find the median
        # standard deviation that might be an approximation to the 'real'
        # standard devation of the data without the bit-flip noise.
        logger.info(
            f"{cid}: More severe handling of images with LIS pixels engaged."
        )
        bf.clean_cube(temp_cube, out_path, width=width, keep=keep)

    if not keep:
        to_del.unlink()

    return


def analyze_cubenorm_stats(
    statsfile: os.PathLike, binning: int, cid=None
) -> tuple:
    with open(statsfile) as csvfile:
        valid_points = list()
        std_devs = list()
        mins = list()
        maxs = list()
        reader = csv.DictReader(csvfile, dialect=isis.cubenormfile.Dialect)
        for row in reader:
            valid_points.append(int(row["ValidPoints"]))
            std_devs.append(float(row["StdDev"]))
            mins.append(int(row["Minimum"]))
            maxs.append(int(row["Maximum"]))

    if cid is None:
        cid = hirise.get_ChannelID_fromfile(statsfile)

    maxvp = max(valid_points)
    logger.info(f"{cid}: Maximum count of valid pixels: {maxvp}")

    # Original note:
    # # Get the median standard deviation value for all columns that have
    # # the maximum valid pixel count
    # #
    # # 2016-12-02 Note: this may not pick the median standard
    # # deviation value but the value from an index 0.95 times the number
    # # of entries in the sorted cubenorm statistics file.
    #
    # That seems to be exactly what it does.  I think the term 'median'
    # in the original (apparently pre-2016) comment is wrong.
    # The variable is called 'facstd' and when it is used below, it refers
    # to this being 'medstd + tol' which indicates that it really is
    # meant to be a factor above the median, which it is.
    std_w_maxvp = list()
    for (vp, std) in zip(valid_points, std_devs):
        if vp == maxvp:
            std_w_maxvp.append(std)

    std_w_maxvp.sort()
    facstd = std_w_maxvp[int((len(std_w_maxvp) - 1) * 0.95)]
    logger.info(
        f"{cid}: 95th percentile standard deviation of all "
        + "columns ({}) that have the ".format(len(std_w_maxvp))
        + "maximum valid pixel count: {}".format(facstd)
    )

    # Original note:
    # # find the minimum of minimums and the maximum of maximums for any
    # # column whose std is less than or equal to $medstd + $tol;
    min_w_maxvp = list()
    max_w_maxvp = list()
    for (vp, std, mi, ma) in zip(valid_points, std_devs, mins, maxs):
        if vp >= (maxvp * 0.9) and std < facstd:
            min_w_maxvp.append(mi)
            max_w_maxvp.append(ma)

    min_w_maxvp.sort()
    max_w_maxvp.sort()

    mindn = min_w_maxvp[int((len(min_w_maxvp) - 1) * 0.05)]
    maxdn = max_w_maxvp[int((len(max_w_maxvp) - 1) * 0.95)]

    if 1 == binning:
        mindn *= 0.7
        maxdn *= 1.3
    elif 2 == binning:
        mindn *= 0.6
        maxdn *= 1.4
    else:
        mindn *= 0.5
        maxdn *= 1.5

    return mindn, maxdn


def HiGainFx(
    cube: os.PathLike,
    outcube: os.PathLike,
    coef_path: os.PathLike,
    version: str,
    keep=False,
) -> None:
    """Perform a Gain-Drift correction on a HiRISE Channel Image."""
    # Eric Eliason (2019 Sept):
    # The HiCal pipeline is apparently double correcting for the
    # Gain Line Drift.   After looking at the source code of the HiCal
    # pipeline, hical program, HiGainFx and the various log files
    # sent to me, I too have come to the same conclusion.  My memory
    # from 10 years ago is starting to come back to me.  Here’s how
    # I remember things:

    # The HiGainFX perl script inserted into the HiCal pipeline was
    # always meant to be a stop-gap measure until a more thorough and
    # robust correction could be developed in the ISIS hical program.
    # Alan D. developed the algorithms for both the stop-gap and
    # full-correction models.  To further complicate things, there
    # were various other intermediary renditions that could be found
    # in the ISIS hicalbeta program but I don’t think hicalbeta was
    # ever used in the HiCal pipeline. I assume hicalbeta eventually
    # became hical.

    warnings.warn(
        "HiGainFx should not be used when Gain Drift correction is "
        "being used in ISIS hical, as HiGainFx is redundant and not "
        "as complete.",
        DeprecationWarning,
    )

    logger.info(HiGainFx.__doc__)
    binning = isis.getkey_k(cube, "Instrument", "Summing")
    ccd = isis.getkey_k(cube, "Instrument", "CcdId")
    chan = isis.getkey_k(cube, "Instrument", "ChannelNumber")

    coef_dir = Path(coef_path)

    if not coef_dir.exists() or not coef_dir.is_dir():
        coef_dir = Path(
            pkg_resources.resource_filename(
                __name__,
                'data/'
            )
        )
        logger.warning(
            "The HiGainFx coefficient directory {} could not be "
            "found, using {} instead.".format(coef_path, coef_dir)
        )

    coef_p = (
        Path(coef_dir)
        / f"HiRISE_Gain_Drift_Correction_Bin{binning}.{version}.csv"
    )

    with open(coef_p) as csvfile:
        reader = csv.DictReader(csvfile, skipinitialspace=True)
        for row in reader:
            if hirise.get_ccdchannel(row["CCD CH"]) == (ccd, chan):
                max_line = row["Max line"]
                a_coef = list()
                for r in (row["R(0)"], row["R(1)"], row["R(2)"]):
                    if r.startswith("-"):
                        # add another '-' for fx syntax
                        a_coef.append(f"-{r}")
                    else:
                        a_coef.append(r)

    eqn = r"\((F1/({0}+({1}*line)+({2}*line*line)))*".format(
        *a_coef
    ) + r"(line<{0}) + (F1*(line>={0})))".format(max_line)

    tfile = outcube.with_suffix(".tempfx.cub")
    isis.fx(f1=cube, to=tfile, mode="CUBES", equation=eqn)
    isis.specadd(tfile, to=outcube, match=cube)
    if not keep:
        tfile.unlink()
    return


def Cubenorm_Filter(
    cubenorm_tab: os.PathLike,
    outfile: os.PathLike,
    pause=False,
    boxfilter=5,
    divide=False,
    chan=None,
) -> tuple:
    """Returns a two-tuple with the standard deviation of the
    filtered average and a boolean indicating whether the maximum
    value of the ValidPoints in *cubenorm_tab* are at one end or
    the other (depending on *chan*).
    """
    logger.debug(
        "Perform a highpass filter on the cubenorm table output of the "
        "columnar average and median values."
    )
    if boxfilter < 3:
        raise ValueError(f"boxfilter={boxfilter} is less than 3")

    if chan is None:
        chan = int(hirise.get_ccdchannel(Path(cubenorm_tab).name)[1])

    # Make a list to receive each column
    valid_points = list()
    averages = list()
    medians = list()
    other_cols = list()
    with open(cubenorm_tab) as csvfile:
        reader = csv.DictReader(csvfile, dialect=isis.cubenormfile.Dialect)
        for row in reader:
            valid_points.append(int(row.pop("ValidPoints")))
            averages.append(float(row.pop("Average")))
            medians.append(float(row.pop("Median")))
            other_cols.append(row)

    zapped = FurrowCheck(valid_points, chan)

    avgflt = Cubenorm_Filter_filter(
        averages,
        boxfilter=boxfilter,
        iterations=50,
        chan=chan,
        pause=pause,
        vpoints=valid_points,
        divide=divide,
    )
    medflt = Cubenorm_Filter_filter(
        medians,
        boxfilter=boxfilter,
        iterations=50,
        chan=chan,
        pause=pause,
        vpoints=valid_points,
        divide=divide,
    )

    with open(outfile, "w") as csvfile:
        writer = isis.cubenormfile.DictWriter(csvfile)
        writer.writeheader()
        for (d, vp, av, md) in zip(other_cols, valid_points, avgflt, medflt):
            d["ValidPoints"] = str(vp)
            d["Average"] = "{:f}".format(av)
            d["Median"] = "{:f}".format(md)
            writer.writerow(d)

    # Calculate the standard deviation value for the filtered average:
    return statistics.stdev(avgflt), zapped


def cut_size(chan: int, length: int) -> collections.namedtuple:
    """Determine the appropriate values for the cut sizes."""
    Cut = collections.namedtuple("Cut", ["left", "right"])
    left = 6
    right = 6

    if 0 == chan:
        if 511 == length:
            left = 40
        elif 255 == length:
            left = 50
    elif 1 == chan:
        if 511 == length:
            right = 40
        elif 255 == length:
            right = 50
    else:
        raise ValueError(f"chan={chan} is not 0 or 1")

    return Cut(left, right)


def Cubenorm_Filter_filter_boxfilter(
    inlist: list, boxfilter: int, iterations=50
) -> list:
    x = copy.deepcopy(inlist)
    ori = copy.deepcopy(inlist)
    hwidth = int(boxfilter / 2)
    frac = [0.25, 0.125]
    for step in range(3):
        for it in range(iterations):
            xflt = copy.deepcopy(x)
            for i, _ in enumerate(x):
                start = i - hwidth
                if start < 0:
                    start = 0
                try:
                    xflt[i] = statistics.mean(
                        filter(lambda y: y != 0, x[start : i + hwidth + 1])
                    )
                except statistics.StatisticsError:
                    xflt[i] = 0
            x = copy.deepcopy(xflt)
        # Zap any columns that are different from the average by more then 25%
        if step < 2:
            for i, (orig, new) in enumerate(zip(ori, x)):
                if (
                    orig != 0
                    and new != 0
                    and (abs(orig - new) / new) > frac[step]
                ):
                    # print(f'orig: {orig}, new: {new}, setting to zero')
                    ori[i] = x[i] = 0
                else:
                    x[i] = orig
                # logger.info(
                # f"frac: {frac[step]}, index: {i}, orig: {orig}, new: {new}, "
                # f"stored: {x[i]}")
    return x


def Cubenorm_Filter_filter(
    inlist: list,
    boxfilter: int,
    iterations: int,
    chan: int,
    pause: bool,
    vpoints: list,
    divide: bool,
) -> list:
    """This performs highpass filtering on the passed list."""
    logger.debug(Cubenorm_Filter_filter.__doc__)

    x = copy.deepcopy(inlist)
    cut = cut_size(chan, len(x))

    # zap the left edge
    x[: cut.left] = [0] * cut.left

    # zap the right edge
    x[(len(x) - cut.right) :] = [0] * cut.right

    # zap the pause point pixels
    if pause and 1024 == len(x):
        for samp, width in zip(util.ch_pause[chan], util.ch_width[chan]):
            zap_slice = util.pause_slicer(samp, width)
            x[zap_slice] = [0] * abs(width)

    # boxfilter
    x = Cubenorm_Filter_filter_boxfilter(x, boxfilter, iterations)

    # Perform the highpass difference of divide the original from lowpass
    maxvp = max(vpoints)
    for i, (orig, vp) in enumerate(zip(inlist, vpoints)):
        if orig != 0 and x[i] != 0 and vp == maxvp:
            if divide:
                x[i] = orig / x[i]
            else:
                x[i] = orig - x[i]
        else:
            x[i] = None

    # Need to patch up any of those None values with neighboring values:
    first_i = (-1, 0)
    patch = (1, -1)

    if x[first_i[chan]] is None:
        x[first_i[chan]] = int(divide)

    if 0 == chan:
        e = reversed(list(enumerate(x)))
    else:
        e = list(enumerate(x))

    for (i, x_val) in e:
        if x_val is None:
            x[i] = x[i + patch[chan]]

    return x


def highlow_destripe(
    high_cube: os.PathLike,
    low_cube: os.PathLike,
    out_cube: os.PathLike,
    conf: dict,
    isisnorm="",
    lnull=True,
    lhrs=True,
    lhis=True,
    llrs=True,
    llis=True,
    keep=False,
) -> None:
    # Perform highpass/lowpass filter vertical destripping
    to_delete = isis.PathSet()
    lpf_cub = to_delete.add(out_cube.with_suffix(".lpf.cub"))
    isis.lowpass(
        low_cube,
        to=lpf_cub,
        line=conf["NoiseFilter_LPF_Line"],
        samp=conf["NoiseFilter_LPF_Samp"],
        minopt="PERCENT",
        replace="NULL",
        minimum=conf["NoiseFilter_LPF_Minper"],
        null=lnull,
        hrs=lhrs,
        his=lhis,
        lrs=llrs,
        lis=llis,
    )

    hpf_cub = to_delete.add(out_cube.with_suffix(".hpf.cub"))
    isis.highpass(
        high_cube,
        to=hpf_cub,
        minopt="PERCENT",
        line=conf["NoiseFilter_HPF_Line"],
        samp=conf["NoiseFilter_HPF_Samp"],
        minimum=conf["NoiseFilter_HPF_Minper"],
    )

    isis.algebra(
        from_=lpf_cub,
        from2=hpf_cub,
        to=out_cube.with_suffix(".cub" + isisnorm),
        operator="ADD",
    )
    if not keep:
        to_delete.unlink()

    return


def getHistVal(histogram: isis.Histogram, conf: dict) -> tuple:
    """Return information about the histogram."""
    lisper = 0
    maxval = None
    if int(histogram["Total Pixels"]) - int(histogram["Null Pixels"]) > 0:
        lisper = (
            int(histogram["Lis Pixels"])
            / (int(histogram["Total Pixels"]) - int(histogram["Null Pixels"]))
            * 100
        )
    cumper = conf["NoiseFilter_HighEnd_Percent"]
    if cumper < 99.0:
        cumper = 99.0

    if lisper > conf["NoiseFilter_Hard_Tolmax"]:
        hard_high_end = conf["NoiseFilter_Hard_HighEnd_Percent"]
        if hard_high_end < 99.9:
            hard_high_end = 99.9
        cumper = hard_high_end

    for row in histogram:
        if float(row.CumulativePercent) > cumper:
            if tuple(isisversion.version_info()[:3]) < (4, 3, 0):
                maxval = float(row.DN)
            else:
                maxval = float(row.MaxExclusive)
            break

    if maxval is None:
        raise ValueError(
            "Did not find a CumulativePercent value greater "
            "than {}".format(cumper)
        )

    return lisper, maxval


def NoiseFilter_noisefilter(
    from_cube: os.PathLike,
    to_cube: os.PathLike,
    flattol: float,
    conf: dict,
    maxval: float,
    tolmin: float,
    tolmax: float,
) -> None:
    """Convenience function for the repeated noisefiltering."""
    isis.noisefilter(
        from_cube,
        to=to_cube,
        flattol=flattol,
        low=conf["NoiseFilter_Minimum_Value"],
        high=maxval,
        tolmin=tolmin,
        tolmax=tolmax,
        sample=conf["NoiseFilter_Noise_Samp"],
        line=conf["NoiseFilter_Noise_Line"],
        toldef="STDDEV",
        replace="NULL",
        lisisnoise=True,
        lrsisnoise=True,
    )


def NoiseFilter_cubenorm_edit(
    in_tab: os.PathLike,
    out_tab: os.PathLike,
    out_tab2: os.PathLike,
    chan: int,
    binning: int,
    conf: dict,
    zapc=False,
) -> None:
    """This function returns nothing, but creates a pair of edited cubenorm
    files.
    """
    # Slightly different values from other function, not entirely sure why.
    # Pause point locations are 1-based pixel numbers, so -1 to get list index.
    # Width values are the number of pixels to affect, including the pause
    # point pixel.
    logger.debug(
        "Zaps the relevant pixels in the cubenorm output and creates an "
        "edited cubenorm file."
    )

    # pause point sample locations:
    ch_pause = {
        0: (1, 252, 515, 778),  # Channel 0
        1: (247, 510, 773, 1024),
    }  # Channel 1

    # Number of pixels to cut from pause point, the sign indicates
    # the direction of cut from the pause point.
    ch_width = {0: (3, 6, 6, 6), 1: (-8, -7, -6, -3)}

    vpnts = list()
    other_cols = list()
    with open(in_tab) as csvfile:
        reader = csv.DictReader(csvfile, dialect=isis.cubenormfile.Dialect)
        for row in reader:
            vpnts.append(int(row.pop("ValidPoints")))
            other_cols.append(row)

    max_vpnts = max(vpnts)
    if max_vpnts <= 0:
        max_vpnts = 1

    # Create a 'unity array'
    # Zap any columns with less then NoiseFilter_Zap_Fraction
    norm = [1] * len(vpnts)
    for i, v in enumerate(vpnts):
        if zapc and v / max_vpnts < float(conf["NoiseFilter_Zap_Fraction"]):
            norm[i] = 0

    # Copy it.
    norm2 = copy.deepcopy(norm)

    # Determine if the pause point pixels need to be zapped
    if binning == 1:
        trigger = NoiseFilter_zaptrigger(
            vpnts,
            max_vpnts,
            conf["NoiseFilter_Nonvalid_Fraction"],
            ch_pause[chan],
            ch_width[chan],
        )

        for samp, width in zip(ch_pause[chan], ch_width[chan]):
            zap_slice = util.pause_slicer(samp, width)
            for i in range(zap_slice.start, zap_slice.stop):
                if trigger:
                    norm[i] = 0
                norm2[i] = 0

    NoiseFilter_cubenorm_writer(out_tab, other_cols, vpnts, norm)
    NoiseFilter_cubenorm_writer(out_tab2, other_cols, vpnts, norm2)

    return


def NoiseFilter_zaptrigger(
    vpnts, max_vpnts, nonvalid_frac, ch_pause, ch_width
) -> bool:
    trigger = False
    for samp, width in zip(ch_pause, ch_width):
        zap_slice = util.pause_slicer(samp, width)
        for i in range(zap_slice.start, zap_slice.stop):
            if vpnts[i] / max_vpnts < nonvalid_frac:
                trigger = True
    if trigger:
        logger.info(f"Fraction of non-valid pixels > {nonvalid_frac}")
        logger.info(f"Pause point pixels will be zapped.")
    return trigger


def NoiseFilter_cubenorm_writer(
    file_path: os.PathLike, other_cols: list, vpnts: list, norm: list
):
    with open(file_path, "w") as csvfile:
        writer = isis.cubenormfile.DictWriter(csvfile)
        writer.writeheader()
        for (d, vp, n) in zip(other_cols, vpnts, norm):
            d["ValidPoints"] = vp
            d["Average"] = n
            d["Median"] = n
            d["StdDev"] = n
            d["Minimum"] = n
            d["Maximum"] = n
            writer.writerow(d)


def NoiseFilter(
    in_cube: os.PathLike,
    output: os.PathLike,
    conf: dict,
    minimum=None,
    maximum=None,
    zapc=False,
    keep=False,
) -> None:
    """NoiseFilter: Perform salt/pepper noise removal."""
    logger.debug(NoiseFilter.__doc__)
    binning = int(isis.getkey_k(in_cube, "Instrument", "Summing"))
    (ccd, chan) = hirise.get_ccdchannel(
        isis.getkey_k(in_cube, "Archive", "ProductId")
    )
    isisnorm = ""
    if minimum is not None and maximum is not None:
        isisnorm = "+SignedWord+{}:{}".format(minimum, maximum)

    to_delete = isis.PathSet()

    h = isis.Histogram(in_cube)
    (LisP, MaxVal) = getHistVal(h, conf)

    cn_tab = to_delete.add(output.with_suffix(".cn.tab"))
    isis.cubenorm(in_cube, stats=cn_tab, format_="TABLE", direction="COLUMN")

    cn2_tab = to_delete.add(output.with_suffix(".cn2.tab"))
    cn3_tab = to_delete.add(output.with_suffix(".cn3.tab"))
    NoiseFilter_cubenorm_edit(
        cn_tab, cn2_tab, cn3_tab, int(chan), binning, conf, zapc
    )

    # Zap the bad columns for the highpass filter
    zap_cub = to_delete.add(output.with_suffix(".zap.cub"))
    isis.cubenorm(
        in_cube,
        to=zap_cub,
        fromstats=cn2_tab,
        statsource="TABLE",
        mode="DIVIDE",
        norm="AVE",
        preserve="FALSE",
    )

    # Zap the bad columns for the lowpass filter
    zap2_cub = to_delete.add(output.with_suffix(".zap2.cub"))
    isis.cubenorm(
        in_cube,
        to=zap2_cub,
        fromstats=cn3_tab,
        statsource="TABLE",
        mode="DIVIDE",
        norm="AVE",
        preserve="FALSE",
    )

    # Perform highpass/lowpass filter vertical destripping
    add_cub = to_delete.add(output.with_suffix(".add.cub"))
    highlow_destripe(
        zap_cub, zap2_cub, add_cub, conf, isisnorm, llis=False, keep=keep
    )

    # Perform the 1st noise filter
    tolmin = float(conf["NoiseFilter_Tolmin"])
    tolmax = float(conf["NoiseFilter_Tolmax"])
    if LisP >= float(conf["NoiseFilter_Hard_Filtering"]):
        tolmin = float(conf["NoiseFilter_Hard_Tolmin"])
        tolmax = float(conf["NoiseFilter_Hard_Tolmax"])
    flattol = float(h["Std Deviation"]) * float(conf["NoiseFilter_Flattol"])
    if flattol < 0.00001:
        flattol = 0.00001

    nf1_cub = to_delete.add(output.with_suffix(".nf1.cub"))
    NoiseFilter_noisefilter(
        add_cub,
        nf1_cub.with_suffix(".cub" + isisnorm),
        flattol,
        conf,
        MaxVal,
        tolmin,
        tolmax,
    )

    # Perform the 2nd noise filter
    nf2_cub = to_delete.add(output.with_suffix(".nf2.cub"))
    NoiseFilter_noisefilter(
        nf1_cub,
        nf2_cub.with_suffix(".cub" + isisnorm),
        flattol,
        conf,
        MaxVal,
        tolmin,
        tolmax,
    )

    # Perform the 3rd noise filter
    nf3_cub = to_delete.add(output.with_suffix(".nf3.cub"))
    NoiseFilter_noisefilter(
        nf2_cub,
        nf3_cub.with_suffix(".cub" + isisnorm),
        flattol,
        conf,
        MaxVal,
        tolmin,
        tolmax,
    )

    # Perform another highpass/lowpass filter now that the
    # data are much cleaner
    add2_cub = to_delete.add(output.with_suffix(".add2.cub"))
    highlow_destripe(
        nf3_cub,
        nf3_cub,
        add2_cub,
        conf,
        isisnorm,
        lnull=False,
        lhrs=False,
        lhis=False,
        llrs=False,
        llis=False,
        keep=keep,
    )

    if ccd.startswith("RED"):
        # Perform LPFZ  filters if we have a RED filter image.
        # For IR and BG filter data, assume that the HiColorNorm pipeline
        # step will interpolate using the BG/RED and IR/RED ratio data.
        lowmin = int(
            int(conf["NoiseFilter_LPFZ_Line"])
            * int(conf["NoiseFilter_LPFZ_Samp"])
            / 3
        )
        lpfz_cub = to_delete.add(output.with_suffix(".lpfz.cub"))
        isis.lowpass(
            add2_cub,
            to=lpfz_cub.with_suffix(".cub" + isisnorm),
            sample=3,
            line=3,
            minopt="COUNT",
            minimum=1,
            filter_="OUTSIDE",
            null=True,
            hrs=False,
            his=True,
            lrs=True,
            lis=True,
        )
        isis.lowpass(
            lpfz_cub,
            to=output.with_suffix(".cub" + isisnorm),
            sample=conf["NoiseFilter_LPFZ_Samp"],
            line=conf["NoiseFilter_LPFZ_Line"],
            minopt="COUNT",
            minimum=lowmin,
            filter_="OUTSIDE",
            null=True,
            hrs=False,
            his=True,
            lrs=True,
            lis=True,
        )
    else:
        to_delete.remove(add2_cub)
        add2_cub.rename(output)

    if not keep:
        to_delete.unlink()
    return


def Hidestripe(
    in_cube: os.PathLike,
    out_cube: os.PathLike,
    binning: int,
    minimum: float,
    maximum: float,
    hidcorr: str,
    line_samples: int,
    keep=False,
) -> float:
    # SignedWord+$HiCal_Normalization_Minimum:$HiCal_Normalization_Maximum
    to_s = "+SignedWord+{}:{}".format(minimum, maximum)
    to_del = isis.PathSet()
    if 1 == binning:
        temp_cub = to_del.add(out_cube.with_suffix(".hd.cub"))
        isis.hidestripe(
            in_cube,
            to=str(temp_cub) + to_s,
            parity="EVEN",
            correction=hidcorr,
        )
        isis.hidestripe(
            temp_cub,
            to=str(out_cube) + to_s,
            parity="ODD",
            correction=hidcorr,
        )
    else:
        lpf_cube = to_del.add(out_cube.with_suffix(".lpf.cub"))
        hpf_cube = to_del.add(out_cube.with_suffix(".hpf.cub"))
        boxsamp = (2 * line_samples) - 1

        isis.lowpass(
            in_cube,
            to=lpf_cube,
            samples=boxsamp,
            lines=3,
            null=False,
            hrs=False,
            his=False,
            lrs=False,
            lis=False,
        )
        isis.highpass(
            in_cube, to=hpf_cube, samples=boxsamp, lines=1, propagate=True
        )
        isis.algebra(
            from_=lpf_cube,
            from2=hpf_cube,
            to=str(out_cube) + to_s,
            operator="ADD",
            a="1.0",
            b="1.0",
        )

    # Standard deviation of difference between cube after noise
    # filter and cube after hidestripe
    diff_cube = to_del.add(out_cube.with_suffix(".diff.cub"))
    isis.algebra(
        from_=out_cube, from2=in_cube, to=diff_cube, operator="subtract"
    )
    diff_pvl = pvl.loads(isis.stats(diff_cube).stdout)["Results"]
    try:
        stddev = float(diff_pvl["StandardDeviation"])
    except (KeyError, ValueError) as err:
        if int(diff_pvl["ValidPixels"]) <= 0:
            raise KeyError(
                "There is no StandardDeviation computed from "
                f"{diff_cube} because there were no ValidPixels. "
                f"Probably because {in_cube} has values outside "
                f"the range: {minimum} to {maximum}."
            )
        else:
            raise err

    if not keep:
        to_del.unlink()

    return stddev


def chan_samp_setup(channel: int, binning: int) -> collections.namedtuple:
    """Returns a named tuple which contains a list and two numbers."""

    if not isinstance(channel, int):
        raise TypeError(
            "channel should be an int, but it "
            "is: {} {}".format(channel, type(channel))
        )

    if not isinstance(binning, int):
        raise TypeError(
            "binning should be an int, but it "
            "is: {} {}".format(binning, type(binning))
        )

    samp = collections.defaultdict(dict)
    # samp[chan][binning]
    samp[0][2] = 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
    samp[1][2] = 512, 511, 510, 509, 508, 507, 506, 505, 504, 503
    samp[0][4] = 1, 2, 3, 4, 5, 6
    samp[1][4] = 256, 255, 254, 253, 252, 251

    ssamp = collections.defaultdict(dict)
    # ssamp[chan][binning]
    ssamp[0][2] = 11
    ssamp[1][2] = 1
    ssamp[0][4] = 7
    ssamp[1][4] = 1

    nsamp = collections.defaultdict(dict)
    # nsamp[chan][binning]
    nsamp[0][2] = 502
    nsamp[1][2] = 502
    nsamp[0][4] = 250
    nsamp[1][4] = 250

    ChanSamp = collections.namedtuple("ChanSamp", ["samp", "ssamp", "nsamp"])

    c = ChanSamp(
        samp[channel][binning],
        ssamp[channel][binning],
        nsamp[channel][binning],
    )
    return c


def furrow_setup(ccd: str, binning: int):
    """Returns the right tuple of furrow."""
    # Expect to call the returned dict like this: d[ccd][binning]
    # Assume each channel for each CCD will have the same threshold
    d = collections.defaultdict(dict)
    d["RED0"][2] = (
        8000,
        8100,
        8700,
        9200,
        9600,
        10000,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED0"][4] = (8000, 9000, 9500, 9900, 9900, 10000)
    d["RED1"][2] = (
        7200,
        7200,
        7800,
        8400,
        9000,
        9500,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED1"][4] = (8000, 8100, 9200, 9600, 9800, 10000)
    d["RED2"][2] = (
        7800,
        7800,
        8400,
        9000,
        9600,
        10000,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED2"][4] = (8000, 8700, 9500, 9800, 9900, 10000)
    d["RED3"][2] = (
        7800,
        8100,
        8300,
        9200,
        9600,
        10000,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED3"][4] = (7900, 9200, 9700, 9900, 10000, 10500)
    d["RED4"][2] = (
        7800,
        7800,
        8300,
        9000,
        9500,
        9900,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED4"][4] = (8000, 8700, 9700, 10000, 10300, 10600)
    d["RED5"][2] = (
        7900,
        8200,
        8600,
        9200,
        9600,
        10000,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED5"][4] = (8000, 9300, 9700, 9900, 10200, 10700)
    d["RED6"][2] = (
        7500,
        7500,
        8100,
        8500,
        9200,
        10000,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED6"][4] = (8000, 8400, 9700, 10000, 10500, 10700)
    d["RED7"][2] = (
        7600,
        8300,
        8900,
        9400,
        9900,
        11000,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED7"][4] = (7700, 9600, 10000, 10200, 11000, 12000)
    d["RED8"][2] = (
        7200,
        7200,
        7900,
        8500,
        9000,
        9400,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED8"][4] = d["RED7"][4]
    d["RED9"][2] = (
        7600,
        8300,
        8600,
        9200,
        9600,
        10000,
        12000,
        12000,
        12000,
        12000,
    )
    d["RED9"][4] = (8000, 8800, 9200, 9400, 9800, 10500)
    d["IR10"][2] = d["RED0"][2]
    d["IR10"][4] = (7600, 8300, 9000, 10000, 10500, 12000)
    d["IR11"][2] = d["RED0"][2]
    d["IR11"][4] = d["IR10"][4]
    d["BG12"][2] = d["RED0"][2]
    d["BG12"][4] = d["IR10"][4]
    d["BG13"][2] = d["RED0"][2]
    d["BG13"][4] = d["IR10"][4]

    return d[ccd][binning]


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Evaluate how the algorithms work.

"""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import math
import subprocess
import sys
import traceback
import json
from pathlib import Path

import numpy as np
from osgeo import gdal_array

import pvl
import kalasiris as isis

import hiproc.util as util
import hiproc.bitflips as bf


# This is only here for temporary testing

dp = Path("/Users/rbeyer/projects/HiRISE/bitflip")


def main():
    try:
        parser = argparse.ArgumentParser(
            description=__doc__,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            parents=[util.parent_parser()],
            conflict_handler="resolve",
        )
        parser.add_argument(
            "-l",
            "--log",
            required=False,
            default=21,
            help="The log level to show for this program, can "
            "be a named log level or a numerical level.",
        )
        parser.add_argument(
            "-w",
            "--width",
            required=False,
            default=5,
            type=int,
            help="The number of medstd widths for bit-flip "
            "cleaning. "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "--saveplot",
            required=False,
            nargs="?",
            default=False,
            const=True,
            help="Saves plot for each area to a file.  If a directory is "
            "provided it will be used to save the plots to, otherwise "
            "the directory of the input file will be used.",
        )

        with open(dp / "dn_ideal.json", "r") as f:
            goal_dn = json.load(f)

        args = parser.parse_args()

        set_logging(args.log, args.logfile)

        atol = 75
        # atol = 1000
        logging.log(21, f"DN must be within {atol} of the goal DN.")

        for cid, v in goal_dn.items():

            fp = dp / v["path"]
            label = pvl.load(fp)
            specialpix = getattr(
                isis.specialpixels, label["IsisCube"]["Core"]["Pixels"]["Type"]
            )
            result = dict()

            print(fp)

            result["img"] = image_area_dn(
                fp,
                specialpix=specialpix,
                width=args.width,
                saveplot=args.saveplot,
            )

            result.update(
                tables_dn(
                    fp,
                    specialpix,
                    v,
                    width=args.width,
                    saveplot=args.saveplot,
                )
            )

            log_lines = [
                f"DN for {fp.name}",
            ]
            mm = ("min", "max")
            # print(result.keys())
            for k in sorted(result.keys()):
                for i in (0, 1):
                    g = v[k][i]
                    r = result[k][i]
                    # print(g)
                    # print(type(g))
                    # print(f"result {r}")
                    # print(type(r))
                    if math.isclose(g, r, abs_tol=atol):
                        # print(f"{k} min is close!")
                        pass
                    else:
                        diff = abs(g - r)
                        # print(diff)
                        # print(type(diff))
                        fdiff = diff / max(g, r)
                        # print(fdiff)
                        # print(type(fdiff))
                        log_lines.append(
                            f"  {k} {mm[i]} is not close, "
                            f"goal: {g}, actual: {r}, diff: {diff}, "
                            f"fdiff: {fdiff:.3}"
                        )
            if len(log_lines) > 1:
                for line in log_lines:
                    # Logs at a level slightly above INFO (20), so the default
                    # just logs this, but selecting INFO also allows underlying
                    # modules to log their INFO-level messages.
                    logging.log(21, line)

        sys.exit(0)
    except subprocess.CalledProcessError as err:
        print("Had an ISIS error:", file=sys.stderr)
        print(" ".join(err.cmd), file=sys.stderr)
        print(err.stdout, file=sys.stderr)
        print(err.stderr, file=sys.stderr)
        sys.exit(1)
    except Exception as err:
        traceback.print_exc(file=sys.stderr)
        print(err, file=sys.stderr)
        sys.exit(1)


def set_logging(i, filename=None) -> None:
    """Sets the log level and basic configuration."""
    if isinstance(i, int):
        log_level = i
    else:
        log_level = getattr(logging, i.upper(), logging.WARNING)

    if filename is None:
        logging.basicConfig(format="%(message)s", level=log_level)
    else:
        logging.basicConfig(
            filename=filename, format="%(message)s", level=log_level,
        )
    return


def image_area_dn(
    in_p: Path, specialpix, width=5, axis=0, saveplot=False,
):
    """ISIS Cube version of clean().

    Please see clean() for argument details.
    """

    image = np.ma.masked_outside(
        gdal_array.LoadFile(str(in_p)), specialpix.Min, specialpix.Max
    )

    # logging.info(f"Bit-flip cleaning Image area.")
    # # These four lines are just informational.
    # img_mean = np.ma.mean(image)
    # img_mode = mstats.mode(image, axis=None)[0][0]
    # d = img_mean - img_mode
    # logging.info(f"Mean: {img_mean}, Mode: {img_mode}, diff: {d}")

    return bf.find_smart_window_from_ma(
        image,
        width=width,
        axis=axis,
        plottitle=(f"{in_p.name} Image Area" if saveplot else None),
        saveplot=(Path(saveplot) / in_p.with_suffix(".bf-image.pdf").name)
        if saveplot
        else False,
    )


def tables_dn(
    in_path: Path, specialpix, keys, width=5, saveplot=False,
):
    rev_area = True if "rev" in keys else False
    mask_area = False
    ramp_area = False
    buffer_area = False
    dark_area = False

    tables_result = dict()

    # print(keys)

    if any((rev_area, mask_area, ramp_area)):
        t_name = "HiRISE Calibration Image"
        HCI_dict = isis.cube.get_table(in_path, t_name)
        cal_vals = np.array(HCI_dict["Calibration"])

        cal_image = np.ma.masked_outside(
            cal_vals, specialpix.Min, specialpix.Max
        )

        # logging.info(f"Bit-flip cleaning Reverse-Clock area.")
        tables_result["rev"] = bf.find_smart_window_from_ma(
            cal_image[:20, :],
            width=width,
            axis=1,
            medstd_limit=200,
            plottitle=(f"{in_path.name} Reverse-Clock" if saveplot else None),
            saveplot=(
                Path(saveplot) / in_path.with_suffix(".bf-revclk.pdf").name
            )
            if saveplot
            else False,
        )

    if any((buffer_area, dark_area)):
        raise NotImplementedError(
            "Bit-flip detection for this area is not yet implemented."
        )

    # print(tables_result)
    return tables_result


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""This module has tests for the bitflips functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import unittest
from pathlib import Path
from unittest.mock import mock_open, patch

import numpy as np
from scipy.signal import find_peaks

import kalasiris as isis

from hiproc import bitflips as bf

# from .utils import resource_check as rc

# # Hardcoding this, but I sure would like a better solution.
# HiRISE_img = Path('test-resources') / 'PSP_010502_2090_RED5_0.img'
# img = HiRISE_img
#
#
# class TestResources(unittest.TestCase):
#     '''Establishes that the test image exists.'''
#
#     def test_resources(self):
#         (truth, test) = rc(img)
#         self.assertEqual(truth, test)


class TestHist(unittest.TestCase):
    def setUp(self):
        HistRow = collections.namedtuple("HistRow", ["DN", "Pixels"])
        self.hist = [
            HistRow(1, 1),  # 0 *
            HistRow(2, 2),  # 1 **
            HistRow(5, 3),  # 2 ***
            HistRow(6, 4),  # 3 ****
            HistRow(7, 3),  # 4 ***
            HistRow(8, 2),  # 5 **
            HistRow(10, 1),  # 6 *
            HistRow(11, 2),  # 7 **
            HistRow(12, 2),  # 8 **
            # Biggest gap here between 12 and 20
            HistRow(20, 1),  # 9 *
        ]

    def test_find_select_idx_name(self):
        self.assertRaises(ValueError, bf.find_select_idx_name, 1, 1, True)
        self.assertEqual(bf.find_select_idx_name(1, 2, True), (-1, "max"))
        self.assertEqual(bf.find_select_idx_name(1, 2, False), (0, "max"))
        self.assertEqual(bf.find_select_idx_name(2, 1, True), (0, "min"))
        self.assertEqual(bf.find_select_idx_name(2, 1, False), (-1, "min"))

    def test_find_minima_index(self):
        pixel_counts = np.fromiter((int(x.Pixels) for x in self.hist), int)
        minima_i, _ = find_peaks(np.negative(pixel_counts))

        self.assertEqual(
            0, bf.find_minima_index(4, 0, minima_i, pixel_counts, default=0)
        )
        self.assertEqual(6, bf.find_minima_index(2, 8, minima_i, pixel_counts))

    def test_find_smart_window(self):

        hist_list = sorted(self.hist, key=lambda x: int(x.DN))
        pixel_counts = np.fromiter((int(x.Pixels) for x in hist_list), int)
        dn = np.fromiter((int(x.DN) for x in hist_list), int)

        self.assertEqual(
            (2, 12),
            bf.find_smart_window(dn, pixel_counts, 1, 12, 6, plot=False),
        )


class TestArrays(unittest.TestCase):
    def test_median_limit(self):
        data = np.array([[1, 2, 3], [3, 4, 5]])

        self.assertEqual(3, bf.median_limit(np.median(data), data))
        self.assertEqual(
            1.5, bf.median_limit(np.median(data), data, limit=2.9)
        )

    def test_medstd(self):
        data = [[1, 3, 3], [9, 5, 6]]

        nomask = np.ma.array(data, mask=np.ma.nomask)
        nm_vp = nomask.count(axis=0)
        nm_sd = np.std(nomask, axis=0)

        mask = np.ma.array(data, mask=[[1, 0, 0], [0, 0, 1]])
        ma_vp = mask.count(axis=0)
        ma_sd = np.std(mask, axis=0)

        self.assertEqual(1.5, bf.median_std(nm_vp, nm_sd))
        self.assertEqual(1, bf.median_std(ma_vp, ma_sd))

    def test_medstd_from_ma(self):
        data = [[1, 3, 3], [9, 5, 6]]

        nomask = np.ma.array(data, mask=np.ma.nomask)
        mask = np.ma.array(data, mask=[[1, 0, 0], [0, 0, 1]])
        self.assertEqual(1.5, bf.median_std_from_ma(nomask))
        self.assertEqual(1, bf.median_std_from_ma(mask))

    # Need to upgrade this test because of find_smart_index
    # def test_clean_array(self):
    #     data = [[1, 3, 3], [9, 5, 6]]

    #     arr = np.ma.array(data)
    #     np.testing.assert_array_equal(
    #         np.ma.masked_outside(arr, 2.9, 5.9), bf.clean_array(arr, width=1)
    #     )

    #     masked = np.ma.array(data, mask=[[1, 0, 0], [0, 0, 1]])
    #     np.testing.assert_array_equal(
    #         np.ma.masked_outside(arr, 2.9, 5.9),
    #         bf.clean_array(masked, width=1),
    #     )

    def test_min_max_ex(self):
        self.assertEqual(bf.min_max_ex(5, 2, 2), (1, 9, 16))
        self.assertEqual(bf.min_max_ex(5, 20, 2), (-35, 45, 20))
        self.assertEqual(bf.min_max_ex(100, 400, 2), (-700, 900, 400))
        self.assertEqual(bf.min_max_ex(100, 401, 2), (-28, 228, 16))

    # Need to upgrade: find_smart_window got complicated, this test is not.
    # def test_find_smart_window_from_ma(self):
    #     data = [[1, 3, 3, 4, 4], [4, 4, 9, 5, 6]]

    #     masked = np.ma.array(data, mask=[[1, 0, 0, 0, 0], [0, 0, 0, 0, 1]])
    #     self.assertEqual(
    #         (3, 9),
    #         bf.find_smart_window_from_ma(masked, width=1, axis=0, plot=False),
    #     )

    def test_mask_lists(self):
        data = [[1, 3, 3], [9, 5, 6]]
        arr = np.ma.array(data, mask=[[1, 0, 0], [0, 0, 1]])

        np.testing.assert_array_equal(
            [[0, 3, 3], [9, 5, 0]],
            bf.apply_special_pixels(arr, isis.specialpixels.UnsignedByte),
        )


class TestMock(unittest.TestCase):
    def setUp(self):
        self.cal = [
            [0, 1000, 1001, 991, 1000, 0],
            [1, 1000, 1002, 992, 1000, 1000],
            [2, 1000, 1003, 993, 1000, 1000],
            [3, 1000, 1004, 994, 1000, 1000],
            [4, 1000, 1005, 995, 1000, 1000],
            [5, 1000, 1011, 990, 1000, 1000],
            [6, 1000, 1010, 990, 1000, 1000],
            [7, 1000, 1010, 990, 1000, 1000],
            [8, 1000, 1010, 990, 1000, 1000],
            [9, 999, 1010, 990, 1001, 1001],
            [10, 1000, 1010, 990, 1000, 1000],
            [11, 1000, 1010, 990, 1000, 1000],
            [12, 1000, 1010, 990, 1000, 1000],
            [13, 1000, 1010, 990, 1000, 1000],
            [14, 1000, 1010, 990, 1000, 1000],
            [15, 1000, 1010, 990, 1000, 1000],
            [16, 1000, 1010, 990, 1000, 1000],
            [17, 1000, 1010, 990, 1000, 1000],
            [18, 1000, 1010, 990, 1000, 1000],
            [19, 1000, 1010, 990, 1000, 1000],
            [20, 1000, 1010, 990, 1000, 1000],
        ]
        self.cleaned = list()
        for row in self.cal:
            # Values less than 3 are less than the specialpixel Min
            # and the reverse-clocked area is only the first 20 lines.
            if 3 <= row[0] < 20:
                self.cleaned.append([0] + row[1:])
            else:
                self.cleaned.append(row)

    def test_clean_cal_tables(self):
        for notimpl in ("mask_area", "ramp_area"):
            self.assertRaises(
                NotImplementedError,
                bf.clean_cal_tables,
                "dummy-in.cub",
                "dummy-out.cub",
                **{notimpl: True}
            )

        masked = np.ma.masked_outside(np.array(self.cal), 3, 65522)

        np.testing.assert_array_equal(
            self.cleaned, bf.clean_cal_tables(masked, 1)
        )

    @patch("hiproc.bitflips.shutil.copy")
    @patch(
        "hiproc.bitflips.pvl.load",
        return_value={
            "IsisCube": {
                "Core": {"Pixels": {"Type": "UnsignedWord"}},
                "Instrument": {"Summing": 1},
            }
        },
    )
    @patch("hiproc.bitflips.open", new_callable=mock_open)
    @patch("hiproc.bitflips.isis.cube.overwrite_table")
    def test_clean_tables_from_cube(self, m_wtable, m_open, m_pvl, m_copy):
        for notimpl in ("mask_area", "ramp_area", "buffer_area", "dark_area"):
            self.assertRaises(
                NotImplementedError,
                bf.clean_tables_from_cube,
                "dummy-in.cub",
                "dummy-out.cub",
                **{notimpl: True}
            )

        with patch(
            "hiproc.bitflips.isis.cube.get_table",
            return_value={"Calibration": self.cal},
        ):
            bf.clean_tables_from_cube(
                Path("dummy-in.cub"),
                Path("dummy-out.cub"),
                width=5,
                rev_area=True,
            )

            # Need to build new test here.
            # self.assertEqual(
            #     self.cleaned, m_wtable.call_args[0][2]["Calibration"]
            # )

    @patch("hiproc.bitflips.clean_tables_from_cube")
    @patch(
        "hiproc.bitflips.pvl.load",
        return_value={
            "IsisCube": {"Core": {"Pixels": {"Type": "UnsignedWord"}}}
        },
    )
    @patch("hiproc.bitflips.isis.mask")
    def test_clean_cube(self, m_mask, m_load, m_clntbl):
        with patch(
            "hiproc.bitflips.gdal_array.LoadFile", return_value=self.cal
        ):
            bf.clean_cube(
                Path("dummy-in.cub"), Path("dummy-out.cub"), keep=True
            )

        self.assertEqual(m_mask.call_args[1]["minimum"], 990)
        self.assertEqual(m_mask.call_args[1]["maximum"], 1010)



#!/usr/bin/env python
"""JitPlot converts the files that HiJitReg creates and displays
a plot which represents them.
"""

# Copyright 2004-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Java programs.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Java
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiColor version 5.4.2 (2020/02/14)
# and on the Java programs:
# - JitPlot.java ($Id: JitPlot.java,v 1.9 2020/02/14 22:46:48 $)
# - JitParser.java ($Id: JitParser.java,v 1.12 2020/02/14 22:46:48 $)
# by Guy McArthur and Michael Wendell as employees of the University of
# Arizona.

import argparse
import logging
import pkg_resources
import statistics

import matplotlib.pyplot as plt
import numpy as np

import pvl
import hiproc.util as util
import hiproc.HiColorInit as hicolor
import hiproc.HiJitReg as hjr

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-c",
        "--conf",
        required=False,
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/HiJitReg.conf'
        ),
        help="Path to the HiJitReg config file.  Defaults to "
             "HiJitReg.conf distributed with the library."
    )
    parser.add_argument(
        "cubes",
        metavar="balance.precolor.cub files",
        nargs="+",
        help="Either one or both sets of RED .balance.cub  and IR/BG "
             ".balance.precolor.cub files. However, that's tedious to type,"
             "so you could just type in *.balance*cub here, and the program "
             "will sort out what it needs."
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    conf = pvl.load(args.conf)

    with util.main_exceptions(args.verbose):
        cubes = list(map(hicolor.HiColorCube, args.cubes))
        (red4, red5, ir10, ir11, bg12, bg13) = hicolor.separate_ccds(cubes)

        plt.ioff()
        fig, axes = plt.subplots(1, 4, sharex=True, sharey=True)

        axes[0].set_ylabel("Line")
        fig.text(0.5, 0.04, "Error Magnitude (pixels)", ha="center", va="center")

        for c, ax in zip([ir10, ir11, bg12, bg13], axes):
            if c is None:
                continue

            logger.info(f"Working on {c}")
            j = hjr.JitterCube(c, conf)
            j.reset()

            (accepted, rejected, smoothed) = filter_and_smooth(j)

            ax.set_title(j.get_ccd())
            size = 5
            if len(accepted) > 0:
                acceptA = np.array(accepted)
                ax.scatter(
                    acceptA[:, 1],
                    acceptA[:, 0],
                    s=size,
                    label="Accepted",
                    facecolors="none",
                    edgecolors="black",
                )
            if len(rejected) > 0:
                rejectedA = np.array(rejected)
                ax.scatter(
                    rejectedA[:, 1],
                    rejectedA[:, 0],
                    c="red",
                    s=size,
                    label="Rejected",
                )
            if len(smoothed) > 0:
                smoothedA = np.array(smoothed)
                ax.scatter(
                    smoothedA[:, 1],
                    smoothedA[:, 0],
                    c="blue",
                    s=size,
                    label="Smoothed",
                )
    fig.suptitle(str(j.get_obsid()))
    bottom, top = plt.ylim()
    plt.ylim(top, bottom)
    plt.xlim(0, 10)
    plt.legend(loc="upper right", bbox_to_anchor=(1, -0.05), ncol=3)
    plt.show()
    return


def filter_and_smooth(j: hjr.JitterCube) -> tuple:
    accepted = list()
    rejected = list()
    smoothed = list()

    for i, cm in enumerate(j.control_measures):
        start = 0
        end = int(i + j["BoxcarLength"] / 2)

        if i > j["BoxcarLength"] / 2:
            start = int(i - j["BoxcarLength"] / 2)

        error_median = statistics.median(
            map(lambda x: x["ErrorMagnitude"], j.control_measures[start:end])
        )

        point = (cm["Row"] * j["LineSpacing"], cm["ErrorMagnitude"])
        smoothed.append((point[0], error_median))

        if (
            abs(cm["ErrorMagnitude"] - error_median) > j["ExcludeLimit"]
            or cm["PointId"] in j.IgnoredPoints
        ):
            rejected.append(point)
            logger.info("Rejecting {}, {}".format(cm["Row"], cm["Column"]))
        else:
            accepted.append(point)

    return accepted, rejected, smoothed



#!/usr/bin/env python
"""HiSlither uses the output of HiJitReg to perform the registration of
color CCDs to red CCDs and output a color cube.

The HiSlither pipeline does a "rubber-sheet stretch" of the IR and
BG to correct for jitter and match the RED based on the control net
output (extension .control.pvl) of HiJitReg. It then stacks the IR,
RED, and BG and on top of one another (in IR-RED-BG order).

The ISIS program slither computes an image transformation that
shifts lines and samples as needed according to the control net
file (extension .control.pvl) created by HiJitReg. It also creates
a text file (extension .slither.txt) containing the transformation
statistics, and a text file (extension .control.csv) from the
smoothing algorithm that checks for "bad" (those that may cause
misregistration) control points. The transformation for each of the
valid control points in a row is averaged together to arrive at a
transformation for the entire row (line). Between rows, the
transformation is interpolated.

The ISIS program hicubeit merges the IR, RED, and BG cubes on top
of one another to create two separate 3-band color products (extension
_COLOR4.cub and _COLOR5.cub).

The ISIS program trim removes the inner overlap from the color halves.



Data Flow
---------
Input Products:

- RED4 and 5 ``balance.cub`` files which are the result of HiccdStitch.
- BG and IR ``precolor.cub`` files which are the result of HiColorInit.
- The regdef.pvl, flat.tab, and control.pvl files created by HiJitReg.

Output Products:

- creates a ``slither.txt`` and ``slither.cub`` file for each of the color
    inputs.
- creates ``COLOR4.cub`` and ``COLOR5.cub`` files which are the multi-band
    aligned images that match RED4 - BG12 - IR10 and RED5 - BG13 - IR11.

"""

# Copyright 2007-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiColor version 5.4.3 (2020/04/28),
# and on the Perl HiSlither program ($Revision: 1.43 $
#                                    $Date: 2020/04/28 16:56:16 $)
# by Guy McArthur as an employee of the University of Arizona.

import argparse
import logging
import math
from datetime import datetime

import kalasiris as isis
import hiproc.util as util
import hiproc.HiColorInit as hicolor
import hiproc.HiJitReg as hjr

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "cubes",
        metavar="RED balance.cub and color balance.precolor.cub files",
        nargs="+",
        help=r"Either one or both sets of RED .balance.cub  and IR/BG "
             r".balance.precolor.cub files. However, that's tedious to type,"
             r"so you could just type in *.balance*cub here, and the program "
             r"will sort out what it needs."
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    with util.main_exceptions(args.verbose):
        HiSlither(args.cubes, keep=args.keep)
    return


def HiSlither(cube_paths: list, keep=False):
    cubes = list(map(hicolor.HiColorCube, cube_paths))
    (red4, red5, ir10, ir11, bg12, bg13) = hicolor.separate_ccds(cubes)

    ccds = list()
    for c in red4, red5, ir10, ir11, bg12, bg13:
        if c is not None:
            ccds.append(str(c))

    logger.info(f"HiSlither start: {', '.join(map(str, ccds))}")

    completed = list()
    if cube_check(red4, ir10, bg12):
        completed.append(str(process_set(red4, ir10, bg12, keep=keep)))

    if cube_check(red5, ir11, bg13):
        completed.append(str(process_set(red5, ir11, bg13, keep=keep)))

    logger.info(f"HiSlither done: {', '.join(completed)}")

    return


def cube_check(
    red: hicolor.HiColorCube, ir: hicolor.HiColorCube, bg: hicolor.HiColorCube
) -> bool:
    if red is None and ir is None and bg is None:
        return False

    if red is None or bg is None:
        # have_ccds = list(filter(lambda x: x is not None, [red, ir, bg]))
        # if len(have_ccds) == 1:
        #     have_str = str(have_ccds[0])
        # else:
        #     have_str = "{} and {}".format(*have_ccds)
        # logger.info(
        #     f"For a slither set, we need at least a RED and BG CCD. "
        #     f"We only have: {have_str}, so no set will be made from these."
        # )
        return False

    return True


def process_set(
    red: hicolor.HiColorCube,
    ir: hicolor.HiColorCube,
    bg: hicolor.HiColorCube,
    keep=False,
):
    if red is None:
        raise TypeError("Expected a RED HiColorCube, but got None.")
    if bg is None:
        raise TypeError("Expected a BG HiColorCube, but got None.")

    logger.info("Beginning HiSlither for {}".format(str(red)))

    temp_token = datetime.now().strftime("HiSlither-%y%m%d%H%M%S")

    if ir is None:
        # If the IR band is missing, it is filled with a "dummy" band
        # consisting entirely of null pixels. This band is created by
        # masking the red band with itself. The dummy band allows for
        # RGB (synthetic B) to be created by HiBeautify even if the
        # IR band is missing.
        logger.info("Missing IR, creating a null band.")
        ir_slither_p = make_dummy_IR(red, bg)
        ir = hicolor.HiColorCube(ir_slither_p)
    else:
        ir_slither_p = run_slither(ir)

    bg_slither_p = run_slither(bg)

    logger.info("Slither finished, merging stacks.")
    logger.info(
        "Creating {}-{}-{} mosaic".format(
            ir.get_ccd(), red.get_ccd(), bg.get_ccd()
        )
    )

    stacked_path = red.path.parent / "{}_COLOR{}.cub".format(
        red.get_obsid(), red.ccdnumber
    )
    tmp_stack = stacked_path.with_suffix(f".{temp_token}.cub")
    isis.hicubeit(ir=ir_slither_p, red=red.path, bg=bg_slither_p, to=tmp_stack)

    trim_args = {"from": tmp_stack, "to": stacked_path}
    # trim 5 px in bin 1, 3 in bin 2, 2 in bin 4
    trim = math.ceil(5 / red.bin)
    if red.ccdnumber == "4":
        trim_args["right"] = trim
    else:
        trim_args["left"] = trim
    isis.trim(**trim_args)

    if not keep:
        tmp_stack.unlink()

    return stacked_path


def get_slither_path(cube):
    return cube.path.parent / "{}.slither.cub".format(str(cube))


def run_slither(cube):
    cnet_path = hjr.JitterCube.get_cnet_path(cube)
    s = get_slither_path(cube)
    t = s.with_suffix(".txt")
    isis.slither(
        cube.path,
        to=s,
        control=cnet_path,
        results=t,
        dir="REVERSE",
        spline="CUBIC",
        interp="BILIN",
    )
    return s


def make_dummy_IR(red: hicolor.HiColorCube, bg: hicolor.HiColorCube):
    bg_slither_path = get_slither_path(bg)
    ir_name = bg_slither_path.name.replace(
        bg.get_ccd(), "IR" + str(int(bg.ccdnumber) - 2)
    )
    ir_path = bg_slither_path.parent / ir_name

    if ir_path.exists():
        raise FileExistsError(
            "{} exists ".format(str(ir_path))
            + "and we don't want to overwrite it."
        )

    ir_ccd = "IR" + str(int(red.ccdnumber) + 6)

    isis.mask(red.path, mask=red.path, to=ir_path, preserve="outside")
    isis.editlab(
        ir_path,
        options="modkey",
        grpname="Instrument",
        keyword="CpmmNumber",
        value=int(red.ccdnumber) + 2,
    )
    isis.editlab(
        ir_path,
        options="modkey",
        grpname="Instrument",
        keyword="CcdId",
        value=ir_ccd,
    )
    isis.editlab(
        ir_path,
        options="modkey",
        grpname="Archive",
        keyword="ProductID",
        value="{}_{}".format(red.get_obsid(), ir_ccd),
    )
    isis.editlab(
        ir_path,
        options="modkey",
        grpname="BandBin",
        keyword="Name",
        value="NearInfrared",
    )
    isis.editlab(
        ir_path,
        options="modkey",
        grpname="BandBin",
        keyword="Center",
        value=900,
        units="NANOMETERS",
    )
    isis.editlab(
        ir_path,
        options="modkey",
        grpname="BandBin",
        keyword="Width",
        value=200,
        units="NANOMETERS",
    )
    return ir_path



#!/usr/bin/env python
"""This module has tests for the HiRISE HiBeautify functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import call, patch
from pathlib import Path

import hiproc.HiBeautify as hbeaut


def getkey(cube, group, key):
    values = {
        "ObservationId": "PSP_010502_2090",
        "ProductId": None,
        "Summing": 2,
        "Lines": 1024,
        "Samples": 1024,
        "Bands": 3,
        "TDI": 64,
        "Center": "(900, 700, 500) <NANOMETERS>",
        "SourceProductId": ("PSP_010502_2090_RED4_0, PSP_010502_2090_RED4_1"),
    }
    return values[key]


class TestHiBeautify(unittest.TestCase):
    # @patch("hiproc.HiColorNorm.ColorCube.get_binning", return_value=2)
    # @patch("hiproc.HiColorNorm.isis.getkey_k", side_effect=getkey)
    # def setUp(self, m_getkey, m_get_binning):
    #     c4 = hcn.ColorCube("dummy/PSP_010502_2090_COLOR4.HiColorNorm")
    #     c5 = hcn.ColorCube("dummy/PSP_010502_2090_COLOR5.HiColorNorm")
    #     self.cubes = [c4, c5]

    @patch("hiproc.HiColorNorm.ColorCube.get_binning", return_value=2)
    @patch("hiproc.HiColorNorm.isis.getkey_k", side_effect=getkey)
    @patch("hiproc.HiBeautify.isis.cubeit_k")
    @patch("hiproc.HiBeautify.isis.algebra")
    @patch("hiproc.HiBeautify.isis.handmos")
    @patch("hiproc.HiBeautify.isis.editlab")
    def test_HiBeautify(
        self,
        m_editlab,
        m_handmos,
        m_algebra,
        m_cubeit_k,
        m_getkey,
        m_get_binning
    ):
        cubes = [
            Path("dummy/PSP_010502_2090_COLOR4.HiColorNorm"),
            Path("dummy/PSP_010502_2090_COLOR5.HiColorNorm")
        ]
        conf = {
            "Beautify": {
                "Synthetic_A_Coefficient": 2,
                "Synthetic_B_Coefficient": 0.3,
            }
        }
        outirb = Path("outirb.cub")
        outrgb = Path("outrgb.cub")
        hbeaut.HiBeautify(cubes, conf, outirb, outrgb, keep=True)

        self.assertEqual(
            m_handmos.call_args_list,
            [
                call(
                    cubes[0],
                    create="Y",
                    mosaic=outirb,
                    nbands=3,
                    nlines=1024,
                    nsamp=2024.0,
                    outband=1,
                    outline=1,
                    outsample=1,
                ),
                call(
                    cubes[1],
                    mosaic=outirb,
                    outband=1,
                    outline=1,
                    outsample=1001,
                ),
            ],
        )

        args, kwargs = m_algebra.call_args
        red = f"{outirb}+2"
        bg = f"{outirb}+3"
        blue = m_algebra.call_args[1]["to"]
        self.assertEqual(m_algebra.call_args[0][0], bg)
        self.assertEqual(m_algebra.call_args[1]["A"], 2)
        self.assertEqual(m_algebra.call_args[1]["B"], 0.3)
        self.assertEqual(m_algebra.call_args[1]["from2"], red)
        self.assertEqual(m_algebra.call_args[1]["op"], "subtract")
        self.assertTrue(str(blue).endswith("_B.cub"))

        self.assertEqual(
            m_cubeit_k.call_args_list, [call([red, bg, blue], to=outrgb)]
        )



#!/usr/bin/env python
"""This module has tests for the HiRISE HiColorNorm functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import call, patch, mock_open
from pathlib import Path

import pvl

import hiproc.hirise as hirise
import hiproc.HiColorNorm as hcn


def getkey(cube, group, key):
    values = {
        "ObservationId": "PSP_010502_2090",
        "ProductId": None,
        "Summing": 2,
        "Lines": 1024,
        "Samples": 1024,
        "Bands": 3,
        "TDI": 64,
        "Center": "(900, 700, 500) <NANOMETERS>",
        "SourceProductId": ("PSP_010502_2090_RED4_0, PSP_010502_2090_RED4_1"),
    }
    return values[key]


class TestColorCube(unittest.TestCase):
    @patch("hiproc.HiColorNorm.ColorCube.get_binning", return_value=2)
    @patch("hiproc.HiColorNorm.isis.getkey_k", side_effect=getkey)
    def setUp(self, m_getkey, m_get_binning):
        self.c = hcn.ColorCube("dummy/PSP_010502_2090_COLOR5")

    def test_init(self):
        self.assertEqual(2, self.c.red_bin)

    def test_get_boxcar_size(self):
        self.assertEqual(1, self.c.get_boxcar_size(2))
        self.assertEqual(3, self.c.get_boxcar_size(3))
        self.assertEqual(5, self.c.get_boxcar_size(5))

    def test_set_crop_lines(self):
        conf = {
            "HiColorNorm": {
                "HiColorNorm_Crop_Top": 1000,
                "HiColorNorm_Crop_Bot": 2000,
            }
        }
        self.assertEqual((1, 1024), self.c.set_crop_lines(conf))

        self.c.lines = 50000
        self.assertEqual((1001, 47000), self.c.set_crop_lines(conf))


class TestHiColorNorm(unittest.TestCase):
    @patch("hiproc.HiColorNorm.ColorCube.get_binning", return_value=2)
    @patch("hiproc.HiColorNorm.isis.getkey_k", side_effect=getkey)
    def setUp(self, m_getkey, m_get_binning):
        c4 = hcn.ColorCube("dummy/PSP_010502_2090_COLOR4")
        c5 = hcn.ColorCube("dummy/PSP_010502_2090_COLOR5")
        self.cubes = [c4, c5]

    @unittest.skip("Tests on a real file.")
    def test_conf_check(self):
        conf_path = Path("data") / "HiColorNorm.conf"
        c = pvl.load(str(conf_path))
        self.assertIsNone(hcn.conf_check(c))

    def test_set_outpath(self):
        self.assertEqual(
            Path("unchanged"), hcn.set_outpath("unchanged", self.cubes)
        )
        self.assertEqual(
            Path("dummy/PSP_010502_2090_COLOR.cub"),
            hcn.set_outpath("_COLOR.cub", self.cubes),
        )
        t_getkey = getkey
        with patch("hiproc.HiColorNorm.ColorCube.get_binning", return_value=2):
            with patch(
                "hiproc.HiColorNorm.isis.getkey_k", side_effect=t_getkey
            ):
                with patch(
                    "hiproc.hirise.get_ObsID_fromfile",
                    return_value=hirise.ObservationID("ESP_060680_3180"),
                ):
                    oddball = hcn.ColorCube("dummy/ESP_060680_3180_COLOR4")
                    newlist = self.cubes + [oddball]
                    self.assertRaises(
                        ValueError, hcn.set_outpath, "unchanged", newlist
                    )

    @patch("hiproc.HiColorNorm.open", new_callable=mock_open)
    def test_FurrowCheck(self, m_open):
        self.assertFalse(hcn.FurrowCheck(self.cubes, None))

        with patch(
            "hiproc.HiColorNorm.json.load", return_value={"zapped": True}
        ):
            self.assertTrue(
                hcn.FurrowCheck(
                    self.cubes, ["dummy_db_path1", "dummy_db_path2"]
                )
            )
        with patch(
            "hiproc.HiColorNorm.json.load", return_value={"zapped": False}
        ):
            self.assertFalse(
                hcn.FurrowCheck(
                    self.cubes, ["dummy_db_path1", "dummy_db_path2"]
                )
            )

    @patch("hiproc.HiColorNorm.isis.handmos")
    def test_make_LR_mosaic(self, m_handmos):
        hcn.make_LR_mosaic("left.cub", "right.cub", 10, "mosaic.cub", 100, 20)
        expected = [
            call(
                "left.cub",
                create="YES",
                mosaic="mosaic.cub",
                nbands=1,
                nlines=100,
                nsamp=20,
                outband=1,
                outline=1,
                outsamp=1,
            ),
            call(
                "right.cub",
                mosaic="mosaic.cub",
                outband=1,
                outline=1,
                outsamp=11,
            ),
        ]
        self.assertListEqual(m_handmos.call_args_list, expected)

    @patch(
        "hiproc.HiColorNorm.csv.DictReader",
        return_value=[{"Average": 4}, {"Average": 5}],
    )
    @patch("hiproc.HiColorNorm.open", new_callable=mock_open)
    @patch("hiproc.HiColorNorm.isis.cubenorm")
    def test_cubenorm_stats(self, m_cubenorm, m_open, m_reader):
        self.assertAlmostEqual(
            0.70710678,
            hcn.cubenorm_stats(
                "crpmos.cub", "mos.cub", "mosnrm.cub", keep=True
            ),
        )
        expected = [
            call(
                "crpmos.cub",
                direction="column",
                format="table",
                norm="average",
                stats=Path("crpmos.cubenorm.txt"),
            ),
            call(
                "mos.cub",
                direction="column",
                format="table",
                fromstats=Path("crpmos.cubenorm.txt"),
                norm="average",
                preserve=True,
                statsource="table",
                to="mosnrm.cub",
            ),
        ]
        self.assertListEqual(m_cubenorm.call_args_list, expected)

    @patch("hiproc.HiColorNorm.isis.handmos")
    @patch("hiproc.HiColorNorm.isis.algebra")
    @patch("hiproc.HiColorNorm.shutil.copyfile")
    def test_make_unfiltered(self, m_copy, m_alg, m_hand):
        self.assertEqual(
            Path("in_UNFILTERED_COLOR4.cub"),
            hcn.make_unfiltered(
                "in_COLOR4.cub", "nrm.cub", "ttoken", "CC", 2, keep=True
            ),
        )
        self.assertEqual(
            m_alg.call_args_list,
            [
                call(
                    "nrm.cub",
                    from2="in_COLOR4.cub+2",
                    operator="MULTIPLY",
                    to=Path("in_COLOR4.ttoken_CC.algebra.cub"),
                )
            ],
        )
        self.assertEqual(
            m_hand.call_args_list,
            [
                call(
                    Path("in_COLOR4.ttoken_CC.algebra.cub"),
                    matchbandbin=False,
                    mosaic=Path("in_UNFILTERED_COLOR4.cub"),
                    outband=2,
                    outline=1,
                    outsample=1,
                )
            ],
        )

    @patch("hiproc.HiColorNorm.isis.lowpass")
    def test_lpfz_filtering(self, m_low):
        hcn.lpfz_filtering("from.cub", "to.cub", 3, 5)
        self.assertListEqual(
            m_low.call_args_list,
            [
                call(
                    "from.cub",
                    HIS=True,
                    HRS=True,
                    LIS=True,
                    LRS=True,
                    filter="OUTSIDE",
                    high=2.0,
                    lines=3,
                    low=0.0,
                    minimum=25,
                    minopt="PERCENTAGE",
                    null=True,
                    samples=5,
                    to="to.cub",
                )
            ],
        )

    @patch("hiproc.HiColorNorm.isis.lowpass")
    def test_lpfz_triplefilter(self, m_low):
        hcn.lpfz_triplefilter("from.cub", "to.cub", keep=True)
        self.assertListEqual(
            m_low.call_args_list,
            [
                call(
                    Path("from.cub"),
                    HIS=True,
                    HRS=True,
                    LIS=True,
                    LRS=True,
                    filter="OUTSIDE",
                    high=2.0,
                    lines=11,
                    low=0.0,
                    minimum=25,
                    minopt="PERCENTAGE",
                    null=True,
                    samples=5,
                    to=Path("from.z1.cub"),
                ),
                call(
                    Path("from.z1.cub"),
                    HIS=True,
                    HRS=True,
                    LIS=True,
                    LRS=True,
                    filter="OUTSIDE",
                    high=2.0,
                    lines=21,
                    low=0.0,
                    minimum=25,
                    minopt="PERCENTAGE",
                    null=True,
                    samples=9,
                    to=Path("from.z2.cub"),
                ),
                call(
                    Path("from.z2.cub"),
                    HIS=True,
                    HRS=True,
                    LIS=True,
                    LRS=True,
                    filter="OUTSIDE",
                    high=2.0,
                    lines=41,
                    low=0.0,
                    minimum=25,
                    minopt="PERCENTAGE",
                    null=True,
                    samples=11,
                    to="to.cub",
                ),
            ],
        )

    @patch("hiproc.HiColorNorm.isis.crop")
    @patch("hiproc.HiColorNorm.isis.mask")
    @patch("hiproc.HiColorNorm.isis.ratio")
    def test_per_color(self, m_rat, m_mask, m_crop):
        mask_p = Path("dummy/PSP_010502_2090_COLOR4.ttoken_IR.mask.cub")
        ratc_p = Path("dummy/PSP_010502_2090_COLOR4.ttoken_IR.ratcrop.cub")
        rati_p = Path("dummy/PSP_010502_2090_COLOR4.ttoken_IR.ratio.cub")

        self.assertTupleEqual(
            (mask_p, ratc_p),
            hcn.per_color(self.cubes[0], "ttoken", "IR", keep=True),
        )
        self.assertEqual(
            m_rat.call_args_list,
            [
                call(
                    den="dummy/PSP_010502_2090_COLOR4+2",
                    num="dummy/PSP_010502_2090_COLOR4+1",
                    to=rati_p,
                )
            ],
        )
        self.assertEqual(
            m_mask.call_args_list,
            [
                call(
                    rati_p,
                    mask=rati_p,
                    maximum=4.0,
                    minimum=0.0,
                    preserve="INSIDE",
                    to=mask_p,
                )
            ],
        )
        self.assertEqual(
            m_crop.call_args_list,
            [call(mask_p, line=None, nlines=None, to=ratc_p)],
        )

    @patch("hiproc.HiColorNorm.isis.handmos")
    @patch("hiproc.HiColorNorm.isis.algebra")
    @patch("hiproc.HiColorNorm.lpfz_triplefilter")
    @patch("hiproc.HiColorNorm.isis.lowpass")
    @patch("hiproc.HiColorNorm.make_unfiltered")
    @patch("hiproc.HiColorNorm.isis.crop")
    @patch("hiproc.HiColorNorm.cubenorm_stats", return_value=3)
    @patch("hiproc.HiColorNorm.make_LR_mosaic")
    def test_per_band(
        self,
        m_LR,
        m_cn,
        m_crop,
        m_unfilt,
        m_low,
        m_3filt,
        m_algebra,
        m_handmos,
    ):
        conf = {
            "HiColorNorm": {
                "HiColorNorm_Crop_Top": 1000,
                "HiColorNorm_Crop_Bot": 2000,
            }
        }
        for c in self.cubes:
            c.set_crop_lines(conf)
        self.assertEqual(
            3,
            hcn.per_band(
                self.cubes,
                Path("out_path.cub"),
                "ttoken",
                "IR",
                True,
                True,
                keep=True,
            ),
        )
        num_cubes = len(self.cubes)
        m_cn.assert_called_once()
        self.assertEqual(num_cubes, m_LR.call_count)
        self.assertEqual(num_cubes, m_crop.call_count)
        self.assertEqual(num_cubes, m_unfilt.call_count)
        self.assertEqual(num_cubes, m_low.call_count)
        self.assertEqual(num_cubes, m_3filt.call_count)
        self.assertEqual(num_cubes, m_algebra.call_count)
        self.assertEqual(num_cubes, m_handmos.call_count)

    @patch("hiproc.HiColorNorm.ColorCube.get_binning", return_value=2)
    @patch("hiproc.HiColorNorm.isis.hiccdstitch")
    @patch("hiproc.HiColorNorm.per_band", return_value=3)
    @patch(
        "hiproc.HiColorNorm.per_color",
        return_value=("CCmask.cub", "CCcrop.cub"),
    )
    @patch("hiproc.HiColorNorm.isis.cubeit")
    @patch("hiproc.HiColorNorm.Path.write_text")
    @patch("hiproc.HiColorNorm.isis.mask")
    @patch("hiproc.HiColorNorm.isis.getkey_k", side_effect=getkey)
    def test_HiColorNorm(
        self, m_getkey, m_mask, m_Pathwrite, m_cubeit, m_pc, m_pb, m_hcs, m_gb
    ):
        conf = {
            "HiColorNorm": {
                "HiColorNorm_Crop_Top": 1000,
                "HiColorNorm_Crop_Bot": 2000,
                "HiColorNorm_Make_Stitch": True,
                "HiColorNorm_NoiseFilter_IR10": True,
                "HiColorNorm_Normalization_Minimum": 0,
                "HiColorNorm_Normalization_Maximum": 16000,
            }
        }
        self.assertEqual(
            (3, 3),
            hcn.HiColorNorm(
                [
                    Path("dummy/PSP_010502_2090_COLOR4"),
                    Path("dummy/PSP_010502_2090_COLOR5")
                ],
                "out_path.cub",
                conf,
                make_unfiltered=True,
                keep=True,
            ),
        )
        self.assertEqual(6, m_mask.call_count)
        self.assertEqual(2, m_cubeit.call_count)
        self.assertEqual(4, m_pc.call_count)
        self.assertEqual(2, m_pb.call_count)



#!/usr/bin/env python
"""Process .dat files from Bob King's processing workflow.

This program is just an adaptor to get Bob's data run through
the algorithms in bitflips.py.

These files are assumed to start with 5 4-byte uint32 values, that
act as a kind of header, and then some number of 4-byte float32
values.
"""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import collections
import os
import sys
from pathlib import Path

import numpy as np

import hiproc.bitflips as bf
import hiproc.util as util

SpecialPixels = collections.namedtuple(
    "SpecialPixels", ["Min", "Null", "Lrs", "Lis", "His", "Hrs", "Max"]
)
# Could also have done:
# import kalasiris.specialpixels.SpecialPixels as SpecialPixels
# but it seemed a little silly to do an import for one namedtuple definition


def main():
    # Set the numpy type for elements in the main data area of the .dat file.
    float_type = np.float32

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[util.parent_parser()],
    )
    parser.add_argument(
        "-o", "--output", required=False, default=".bitflip.dat"
    )
    parser.add_argument(
        "-w",
        "--width",
        required=False,
        default=5,
        help="The number of medstd widths for bit-flip " "cleaning.",
    )
    parser.add_argument(
        "-r",
        "--replacement",
        required=False,
        default=float_type(0),
        type=float_type,
        help="By default, the program will replace "
        "identified pixels with an appropriate NULL data "
        "value, but if provided this value will be used "
        "instead.",
    )
    parser.add_argument(
        "-p",
        "--plot",
        required=False,
        action="store_true",
        help="Displays plot for each area.",
    )
    parser.add_argument(
        "-n",
        "--dryrun",
        required=False,
        action="store_true",
        help="Does not produce a cleaned output file.",
    )
    parser.add_argument("file", help="A .dat file to clean.")

    args = parser.parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    out_p = util.path_w_suffix(args.output, args.file)

    clean(
        args.file,
        out_p,
        args.replacement,
        width=args.width,
        plot=args.plot,
        dryrun=args.dryrun,
    )

    sys.exit(0)


def clean(
    in_path: os.PathLike,
    out_path: os.PathLike,
    replacement,
    width=5,
    plot=False,
    dryrun=False,
):
    """The file at *out_path* will be the result of running bit-flip
    cleaning of the file at *in-path*.

    **WARNING**: at this time, in_path is checked to conform to being
    a series of 20-line observations, and if so, then reverse-clock
    cleaning will be applied.

    The *replacement* value is used for two things.  Firstly, the
    dtype of replacement is used to interpret the data in *in_path*
    after the header.  Secondly, pixels that are identified
    as being beyond the allowable DN window (which may be defined
    differently for each of the image areas), will be set to this
    value.
    """

    in_p = Path(in_path)
    out_p = Path(out_path)

    header_count = 5
    header = np.fromfile(in_p, dtype=np.uint32, count=header_count, sep="")
    # Looks like: [   2  256   20    4 5120]
    # cols = header[1]
    # rows = header[2]

    if header[2] != 20:
        raise ValueError(
            "The file did not contain 20-line observations, "
            "and is not a set of reverse-clock data."
        )
    # print(header)

    a = np.fromfile(
        in_p,
        dtype=replacement.dtype,
        count=-1,
        sep="",
        offset=(header_count * replacement.dtype.itemsize),
    )

    if a.shape[0] % (header[1] * header[2]) > 0:
        raise ValueError(
            f"The file had {a.shape} float values that were "
            f"not evenly divisible by {header[1]} x {header[2]}."
        )

    obs_num = int(a.shape[0] / header[2] / header[1])
    rev_clocks = a.reshape((obs_num, header[2], header[1]))
    # print(rev_clocks)
    # print(rev_clocks.shape)

    # Special Pixels
    # ISIS has 5 special pixel 'values', not all of which are needed
    # for these calculations, but the called functions expect the
    # structure, so we must create one to use.
    #
    # For reasons that are unknown to me, there is one Special Pixel
    # situation for images that have no LUT applied, and a different
    # one for those that have.  To understand the differences, please
    # see the specialpixels() function of the LUT_Table class in the
    # img.py file of this distro.  For this data, we will use these
    # values, and allow the user to set the Null value.
    specialpix = SpecialPixels(
        Min=1, Null=replacement, Lrs=0, Lis=0, His=16383, Hrs=65535, Max=16382
    )
    if specialpix.Min < specialpix.Null < specialpix.Max:
        raise ValueError(
            "The replacement value is numerically between "
            f"the valid minimum ({specialpix.Min}) and "
            f"the valid maximum ({specialpix.Max})."
        )

    for obs in range(rev_clocks.shape[0]):
        this_slice = np.s_[obs, :, :]
        r = rev_clocks[this_slice]

        # print(r)

        if r[r == 0].size:
            percent = (r[r == 0].size / r.size) * 100
            desc = f"Observation {obs} has {r[r == 0].size} zeros, {percent}%."

            if percent > 40:
                print(f"{desc} Too many zeros, will not process.")
                continue

            print(desc)
            r_masked = np.ma.masked_invalid(
                np.ma.masked_outside(r, specialpix.Min, specialpix.Max)
            )

            # print(r_masked)
            r_clean = bf.clean_array(
                r_masked,
                width=width,
                axis=1,
                plot=(f"Observation {obs} Reverse-Clock" if plot else False),
            )
            rev_clocks[this_slice] = unmask(r_clean, specialpix)
    if not dryrun:
        with open(out_p, mode="wb") as f:
            f.write(header.tobytes())
            f.write(rev_clocks.tobytes())
    return


def unmask(array: np.ma, specialpix) -> np.ma:
    """First remove the mask from any existing NaN or inf pixels,
    then apply the *specialpix* to any remaining masked pixels.
    """
    array.mask = np.logical_xor(
        np.ma.masked_invalid(array.data).mask, array.mask
    )
    return bf.apply_special_pixels(array, specialpix)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Convert one of Alan's MDR files (in DN or I/F) to a cube file (in I/F)."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This program is based on equations in make_IOF.pro, by Alan Delamere,
# 2019-05-22, and on equations in the public domain ISIS program hical.

import argparse
import logging
import pkg_resources
import shutil
import statistics
import sys
from pathlib import Path
from osgeo import gdal

import pvl
import spiceypy

import hiproc.hirise as hirise
import hiproc.util as util
import kalasiris as isis

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-o", "--output", required=False, default=".mdr.iof.cub"
    )
    parser.add_argument("-e", "--edr", required=True)
    parser.add_argument(
        "-c",
        "--conf",
        required=False,
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/hical.pipelines.conf'
        ),
    )
    parser.add_argument("mdr", metavar="MDR_file")
    parser.add_argument(
        "-l",
        "--log",
        required=False,
        default="WARNING",
        help="The log level to show for this program, can "
        "be a named log level or a numerical level.",
    )
    parser.add_argument(
        "-k",
        "--keep",
        required=False,
        default=False,
        action="store_true",
        help="Normally, the program will clean up any "
        "intermediary files, but if this option is given, it "
        "won't.",
    )

    args = parser.parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    edr_path = Path(args.edr)
    mdr_path = Path(args.mdr)

    to_del = isis.PathSet()

    h2i_path = to_del.add(edr_path.with_suffix(".hi2isis.cub"))

    out_path = util.path_w_suffix(args.output, edr_path)

    # The first thing Alan's program did was to crop the image down to only the
    # 'imaging' parts.  We're not doing that so the resultant file has a
    # geometry similar to what comes out of ISIS hical.

    # Convert the EDR to a cube file
    isis.hi2isis(edr_path, to=h2i_path)

    # Convert Alan's MDR to a cube file
    mdr_cub_path = to_del.add(mdr_path.with_suffix(".alan.cub"))
    logger.info(f"Running gdal_translate {mdr_path} -of ISIS3 {mdr_cub_path}")
    gdal.Translate(str(mdr_cub_path), str(mdr_path), format="ISIS3")

    h2i_s = int(isis.getkey_k(h2i_path, "Dimensions", "Samples"))
    h2i_l = int(isis.getkey_k(h2i_path, "Dimensions", "Lines"))
    mdr_s = int(isis.getkey_k(mdr_cub_path, "Dimensions", "Samples"))
    mdr_l = int(isis.getkey_k(mdr_cub_path, "Dimensions", "Lines"))

    if h2i_s != mdr_s:
        label = pvl.load(str(h2i_path))
        hirise_cal_info = get_one(
            label, "Table", "HiRISE Calibration Ancillary"
        )

        buffer_pixels = get_one(hirise_cal_info, "Field", "BufferPixels")[
            "Size"
        ]
        dark_pixels = get_one(hirise_cal_info, "Field", "DarkPixels")["Size"]
        rev_mask_tdi_lines = hirise_cal_info["Records"]

        if h2i_s + buffer_pixels + dark_pixels == mdr_s:
            logger.info(
                f"The file {mdr_cub_path} has "
                f"{buffer_pixels + dark_pixels} more sample pixels "
                f"than {h2i_path}, assuming those are dark and "
                "buffer pixels and will crop accordingly."
            )
            if h2i_l + rev_mask_tdi_lines != mdr_l:
                logger.critical(
                    'Even assuming this is a "full" channel '
                    "image, this has the wrong number of lines. "
                    f"{mdr_cub_path} should have "
                    f"{h2i_l + rev_mask_tdi_lines}, but "
                    f"has {mdr_l} lines. Exiting"
                )
                sys.exit()
            else:
                crop_path = to_del.add(mdr_cub_path.with_suffix(".crop.cub"))
                # We want to start with the next pixel (+1) after the cal
                # pixels.
                isis.crop(
                    mdr_cub_path,
                    to=crop_path,
                    sample=buffer_pixels + 1,
                    nsamples=h2i_s,
                    line=rev_mask_tdi_lines + 1,
                )
                mdr_cub_path = crop_path
                mdr_l = int(isis.getkey_k(mdr_cub_path, "Dimensions", "Lines"))

        else:
            logger.critical(
                f"The number of samples in {h2i_path} ({h2i_s}) "
                f"and {mdr_cub_path} ({mdr_s}) are different. "
                "Exiting."
            )
            sys.exit()

    if h2i_l != mdr_l:
        logger.critical(
            f"The number of lines in {h2i_path} ({h2i_l}) "
            f"and {mdr_cub_path} ({mdr_l}) are different. "
            "Exiting."
        )
        sys.exit()

    # Convert the EDR to the right bit type for post-HiCal Pipeline:
    h2i_16b_p = to_del.add(h2i_path.with_suffix(".16bit.cub"))
    isis.bit2bit(
        h2i_path,
        to=h2i_16b_p,
        bit="16bit",
        clip="minmax",
        minval=0,
        maxval=1.5,
    )
    shutil.copyfile(h2i_16b_p, out_path)

    # If it is a channel 1 file, Alan mirrored it so that he could process
    # the two channels in an identical way (which we also took advantage
    # of above if the buffer and dark pixels were included), so we need to
    # mirror it back.
    cid = hirise.get_ChannelID_fromfile(h2i_16b_p)
    if cid.channel == "1":
        mirror_path = to_del.add(mdr_cub_path.with_suffix(".mirror.cub"))
        isis.mirror(mdr_cub_path, to=mirror_path)
        mdr_cub_path = mirror_path

    # Is the MDR in DN or I/F?
    maximum_pxl = float(
        pvl.loads(isis.stats(mdr_cub_path).stdout)["Results"]["Maximum"]
    )
    if maximum_pxl < 1.5:
        logger.info("MDR is already in I/F units.")
        mdr_16b_p = to_del.add(mdr_cub_path.with_suffix(".16bit.cub"))
        isis.bit2bit(
            mdr_cub_path,
            to=mdr_16b_p,
            bit="16bit",
            clip="minmax",
            minval=0,
            maxval=1.5,
        )
        isis.handmos(mdr_16b_p, mosaic=out_path)
    else:
        logger.info("MDR is in DN units and will be converted to I/F.")

        fpa_t = statistics.mean(
            [
                float(
                    isis.getkey_k(
                        h2i_16b_p, "Instrument", "FpaPositiveYTemperature"
                    )
                ),
                float(
                    isis.getkey_k(
                        h2i_16b_p, "Instrument", "FpaNegativeYTemperature"
                    )
                ),
            ]
        )
        print(f"fpa_t {fpa_t}")

        conf = pvl.load(args.conf)

        tdg = t_dep_gain(get_one(conf["Hical"], "Profile", cid.ccdname), fpa_t)
        suncorr = solar_correction()
        sclk = isis.getkey_k(
            h2i_16b_p, "Instrument", "SpacecraftClockStartCount"
        )
        target = isis.getkey_k(h2i_16b_p, "Instrument", "TargetName")
        suncorr = solar_correction(sunDistanceAU(sclk, target))
        sed = float(
            isis.getkey_k(h2i_16b_p, "Instrument", "LineExposureDuration")
        )
        zbin = get_one(conf["Hical"], "Profile", "GainUnitConversion")[
            "GainUnitConversionBinFactor"
        ]

        # The 'ziof' name is from the ISIS HiCal/GainUnitConversion.h, it is a
        # divisor in the calibration equation.
        print(f"zbin {zbin}")
        print(f"tdg {tdg}")
        print(f"sed {sed}")
        print(f"suncorr {suncorr}")
        ziof = zbin * tdg * sed * 1e-6 * suncorr
        eqn = f"\(F1 / {ziof})"  # noqa W605

        mdriof_p = to_del.add(mdr_cub_path.with_suffix(".iof.cub"))
        to_s = "{}+SignedWord+{}:{}".format(mdriof_p, 0, 1.5)
        isis.fx(f1=mdr_cub_path, to=to_s, equ=eqn)

        isis.handmos(mdriof_p, mosaic=out_path)

    if not args.keep:
        to_del.unlink()


def solar_correction(au=1.498) -> float:
    # au = 1.498  # Placeholder from Alan
    s = 1.5 / au  # Not sure about this, comes from the ISIS code
    return s * s


def sunDistanceAU(time: str, target: str) -> float:
    """Returns distance in AU between Sun and observed body from MRO."""

    base_kernel_path = Path(isis.environ["ISIS3DATA"]) / "base" / "kernels"
    lsk = sorted(Path(base_kernel_path / "lsk").glob("naif*.tls"))[-1]
    pck = sorted(Path(base_kernel_path / "spk").glob("de*.bsp"))[-1]
    sat = sorted(Path(base_kernel_path / "spk").glob("mar*.bsp"))[-1]

    sclk = sorted(
        Path(
            Path(isis.environ["ISIS3DATA"]) / "mro" / "kernels" / "sclk"
        ).glob("MRO_SCLKSCET.*.65536.tsc")
    )[-1]

    spiceypy.furnsh([str(lsk), str(pck), str(sat), str(sclk)])

    et = spiceypy.scs2e(-74999, time)

    targ = target.lower()
    if targ == "sky" or targ == "cal" or targ == "phobos" or targ == "deimos":
        targ = "mars"

    (sunv, lt) = spiceypy.spkpos(targ, et, "J2000", "LT+S", "sun")

    sunkm = spiceypy.vnorm(sunv)
    # Return in AU units
    return sunkm / 1.49597870691e8


def t_dep_gain(profile: dict, t: float) -> float:
    """Given the profile, and the FPA temperature in C,
    calculate the temperature dependent gain."""
    # Equivalent to getTempDepGain() in ISIS HiCal/GainUnitConversion.h
    # These equations are really g * (1 + t - baseT) * Q * absgainTDI
    # where these variables are from the hical.pipelines.conf file.
    #   g = FilterGainCorrection (DN/s) - Alan's numbers don't match
    #   baseT = IoverFbasetemperature (deg C)  - Alan's numbers are the same
    #   Q = QEpercentincreaseperC (1/deg C) - Alan's numbers are the same
    #   absgainTDI = AbsGain_TDI128 (? units) - Alan's numbers don't match
    #
    # These are Alan's numbers from the clean11 era.  Need to hunt down
    # authoritative values and/or verify hical.pipelines.conf which had a
    # syntax error.
    # zgain = dict(RED=157709797, IR=56467454, BG=115121269)
    # baseT = dict(RED=18.9, IR=18.9, BG=18.9)
    # QEpcntC = dict(RED=0.0005704, IR=0.002696, BG=0.00002295)
    # absgainTDI = dict(RED=6.37688, IR=6.99017, BG=7.00042)
    zgain = profile["FilterGainCorrection"]
    baseT = profile["IoverFbasetemperature"]
    QEpcntC = profile["QEpercentincreaseperC"]
    absgainTDI = profile["AbsGain_TDI128"]

    return zgain * (1 + (t - baseT) * QEpcntC * absgainTDI)


def get_one(conf, thing: str, name: str) -> dict:
    """Return the item for the named thing (of which there are multiple)."""
    for it in conf.getlist(thing):
        if it["Name"] == name:
            return it
